"""
Dynamic content.
"""

from __future__ import annotations

from abc import abstractmethod
from asyncio import gather
from collections.abc import Iterable, Mapping
from typing import TYPE_CHECKING, Any, Self, final, override

from betty.ancestry.has_notes import HasNotes
from betty.content import (
    Content,
    ContentDefinition,
    ContentManufacturer,
    build,
)
from betty.data import Data, Sample
from betty.data.aggregate.record.object import ObjectDefinition
from betty.data.str import StrDefinition
from betty.locale.localizable.gettext import _
from betty.locale.localizable.property import LocalizableProperty
from betty.locale.localize import resolve_localized
from betty.media_type import MediaType
from betty.media_type.media_types import PLAIN_TEXT
from betty.plugin.data.property import PluginManufacturerSequenceProperty
from betty.property import Optional, Property
from betty.sample import Size
from betty.service.factory import DataManufacturable, Manufacturable
from betty.service.requirement.project import require_project

if TYPE_CHECKING:
    from collections.abc import MutableSequence

    from betty.document import Document
    from betty.jinja import Environment
    from betty.locale.localizable import ResolvableLocalizable
    from betty.plugin.factory import ResolvablePluginManufacturerSequence
    from betty.project import Project
    from betty.render import RenderDispatcher


@final
@ObjectDefinition(
    label=_("Rendered content configuration"),
    samples=[
        lambda: Sample(
            RenderConfiguration("Hello, world!"), label="Minimal", size=Size.MINIMAL
        )
    ],
)
class RenderConfiguration(Data):
    """
    Configuration for :py:class:`betty.content.contents.Render`.

    .. data:: betty.content.contents:RenderConfiguration
    """

    content = LocalizableProperty(label=_("Content"))
    media_type = Property(MediaType, default=lambda: PLAIN_TEXT, omit_load=True)

    def __init__(
        self, /, content: ResolvableLocalizable, media_type: MediaType = PLAIN_TEXT
    ):
        super().__init__()
        self.content = content
        self.media_type = media_type


@ContentDefinition("render", label=_("Rendered content"))
class Render(DataManufacturable[RenderConfiguration], Content):
    """
    .. plugin:: content:render.
    """

    def __init__(
        self,
        *,
        content: ResolvableLocalizable,
        renderer: RenderDispatcher,
        media_type: MediaType = PLAIN_TEXT,
    ):
        self._content = content
        self._media_type = media_type
        self._renderer = renderer

    @override
    @classmethod
    def new_data_cls(cls) -> type[RenderConfiguration]:
        return RenderConfiguration

    @override
    @classmethod
    @require_project
    async def new(cls, project: Project, data: RenderConfiguration, /) -> Self:
        return cls(
            content=data.content,
            media_type=data.media_type,
            renderer=await project.renderer,
        )

    @override
    async def build(self, *, document: Document) -> str | None:
        return await self._renderer.render(
            resolve_localized(self._content, localizer=document.localizer),
            self._media_type,
        )


type TemplateBuild = (
    str | Iterable[str] | tuple[str | Iterable[str], Mapping[str, Any]] | None
)


class Template(Content):
    """
    Build content by rendering a Jinja2 template.
    """

    def __init__(self, *args: Any, jinja: Environment, **kwargs: Any):
        super().__init__(*args, **kwargs)
        self._jinja = jinja

    @override
    async def build(self, *, document: Document) -> str | None:
        config = await self.build_template(document)
        if config is None:
            return None
        templates: MutableSequence[str]
        if isinstance(config, str):
            templates = [config]
            data = {}
        elif isinstance(config, tuple):
            templates = [config[0]] if isinstance(config[0], str) else config[0]  # ty:ignore[invalid-assignment]
            data = config[1]
        else:
            templates = config  # ty:ignore[invalid-assignment]
            data = {}
        assert templates, "At least one template must be specified"
        rendered_content = (
            await self._jinja.select_template(templates).render_async(
                document=document,
                **data,  # ty:ignore[invalid-argument-type]
            )
        ).strip()
        if rendered_content:
            return rendered_content
        return None

    def _resolve_templates(self, templates: str | Iterable[str]) -> Iterable[str]:
        if isinstance(templates, str):
            return [templates]
        return templates

    @abstractmethod
    async def build_template(self, document: Document) -> TemplateBuild:
        """
        Build template data.

        Return a template name, a tuple of a template name and template date to render it. Return ``None`` to prevent
        anything from being rendered at all.
        """


@ContentDefinition("notes", label=_("Notes"))
class Notes(Template, Manufacturable):
    """
    .. plugin:: content:notes.
    """

    @override
    @classmethod
    @require_project
    async def new(cls, project: Project, /) -> Self:
        return cls(jinja=await project.jinja)

    @override
    async def build_template(self, document: Document) -> TemplateBuild:
        if isinstance(document.resource, HasNotes):
            return "component/notes.html.j2", {"notes": document.resource.notes}
        return None


@final
@ObjectDefinition(
    label=_("Box configuration"),
    samples=[
        lambda: Sample(BoxConfiguration([]), label="Minimal", size=Size.MINIMAL),
        lambda: Sample(
            BoxConfiguration(
                [ContentManufacturer(Render, RenderConfiguration("Hello, world!"))],
                min_height="100px",
                max_height="1000px",
                height="500px",
                min_width="100px",
                max_width="1000px",
                width="500px",
            ),
            label="Full",
            size=Size.FULL,
        ),
    ],
)
class BoxConfiguration(Data):
    """
    Configuration for :py:class:`betty.content.contents.Box`.

    .. data:: betty.content.contents:BoxConfiguration
    """

    content = PluginManufacturerSequenceProperty[ContentDefinition, Content](
        ContentManufacturer, label=_("Content")
    )
    """
    The content within this box.
    """

    min_height = Optional(Property(StrDefinition(label=_("Minimum height"))))
    max_height = Optional(Property(StrDefinition(label=_("Maximum height"))))
    height = Optional(Property(StrDefinition(label=_("Height"))))
    min_width = Optional(Property(StrDefinition(label=_("Minimum width"))))
    max_width = Optional(Property(StrDefinition(label=_("Maximum width"))))
    width = Optional(Property(StrDefinition(label=_("Width"))))

    def __init__(
        self,
        content: ResolvablePluginManufacturerSequence[ContentDefinition, Content],
        *,
        min_height: str | None = None,
        max_height: str | None = None,
        height: str | None = None,
        min_width: str | None = None,
        max_width: str | None = None,
        width: str | None = None,
    ):
        super().__init__()
        self.content = content
        self.min_height = min_height
        self.max_height = max_height
        self.height = height
        self.min_width = min_width
        self.max_width = max_width
        self.width = width


@final
@ContentDefinition("box", label=_("Box"))
class Box(Template, DataManufacturable[BoxConfiguration]):
    """
    .. plugin:: content:box.
    """

    def __init__(
        self,
        /,
        content: Iterable[Content],
        *,
        jinja: Environment,
        min_height: str | None = None,
        max_height: str | None = None,
        height: str | None = None,
        min_width: str | None = None,
        max_width: str | None = None,
        width: str | None = None,
    ):
        super().__init__(jinja=jinja)
        self._content = tuple(content)
        self._min_height = min_height
        self._max_height = max_height
        self._height = height
        self._min_width = min_width
        self._max_width = max_width
        self._width = width

    @override
    @classmethod
    def new_data_cls(cls) -> type[BoxConfiguration]:
        return BoxConfiguration

    @override
    @classmethod
    @require_project
    async def new(cls, project: Project, data: BoxConfiguration, /) -> Self:
        content, jinja = await gather(
            gather(
                *map(
                    project.factory.new,
                    map(ContentManufacturer.resolve, data.content),
                )
            ),
            project.jinja,
        )
        return cls(
            content=content,
            min_height=data.min_height,
            max_height=data.max_height,
            height=data.height,
            min_width=data.min_width,
            max_width=data.max_width,
            width=data.width,
            jinja=jinja,
        )

    @override
    async def build_template(self, document: Document) -> TemplateBuild:
        content = await build(document, self._content)
        if content is None:
            return None
        return "component/box.html.j2", {
            "box_content": content,
            "box_min_height": self._min_height,
            "box_max_height": self._max_height,
            "box_height": self._height,
            "box_min_width": self._min_width,
            "box_max_width": self._max_width,
            "box_width": self._width,
        }



"""
Linked data for the date API.
"""

from betty.date import Date, DateRange, _dump_date_iso8601
from betty.json.linked_data import dump_context
from betty.portable import PortableMapping


def dump_linked_data_for_date(
    date: Date, *, context_definition: str | None = None
) -> PortableMapping:
    """
    Dump a date to linked data.
    """
    portable: PortableMapping = {
        "fuzzy": date.fuzzy,
    }
    if date.year:
        portable["year"] = date.year
    if date.month:
        portable["month"] = date.month
    if date.day:
        portable["day"] = date.day
    if date.comparable:
        portable["iso8601"] = _dump_date_iso8601(date)
        # Set a single term definition because JSON-LD does not let us apply multiple
        # for the same term (key).
        if context_definition:
            dump_context(portable, iso8601=context_definition)
    return portable


def dump_linked_data_for_date_range(
    date_range: DateRange,
    *,
    start_context_definition: str | None = None,
    end_context_definition: str | None = None,
) -> PortableMapping:
    """
    Dump a date range to linked data.
    """
    return {
        "start": dump_linked_data_for_date(
            date_range.start, context_definition=start_context_definition
        )
        if date_range.start
        else None,
        "end": dump_linked_data_for_date(
            date_range.end, context_definition=end_context_definition
        )
        if date_range.end
        else None,
    }



"""
The Assets API.
"""

from __future__ import annotations

from abc import ABC, abstractmethod
from asyncio import to_thread
from contextlib import suppress
from os import walk
from pathlib import Path
from typing import TYPE_CHECKING, override

from betty.concurrent import AsynchronizedLock
from betty.typing import threadsafe

if TYPE_CHECKING:
    from collections.abc import AsyncIterable, Iterable, Mapping, Sequence


class AssetError(Exception):
    """
    Raised for asset API errors.
    """


class UnknownAsset(AssetError):
    """
    Raised when a requested asset cannot be found.
    """

    def __init__(self, path: Path, assets_directory_paths: Iterable[Path], /):
        super().__init__(
            f"Asset {path} cannot be found in any of: {', '.join(map(str, assets_directory_paths))}"
        )


@threadsafe
class AssetRepository(ABC):
    """
    Manages a set of assets.

    This repository unifies several directory paths on disk, overlaying them on
    each other. Paths added later act as fallbacks, e.g. earlier paths have priority.
    """

    @property
    @abstractmethod
    def directories(self) -> Sequence[Path]:
        """
        The paths to the individual virtual layers.
        """

    @abstractmethod
    def walk(self, directory: Path | None = None, /) -> AsyncIterable[Path]:
        """
        Get virtual paths to available assets.

        :param directory: If given, only asses under the directory are returned.
        """

    @abstractmethod
    async def get(self, path: Path, /) -> Path:
        """
        Get the path to a single asset file.

        :param path: The virtual asset path.
        :return: The path to the actual file on disk.
        """


class ProxyAssetRepository(AssetRepository):
    """
    Provides assets from upstream repositories.
    """

    def __init__(self, *upstreams: AssetRepository):
        self._upstreams = upstreams

    @override
    @property
    def directories(self) -> Sequence[Path]:
        return [path for upstream in self._upstreams for path in upstream.directories]

    @override
    async def walk(self, directory: Path | None = None, /) -> AsyncIterable[Path]:
        seen = set()
        for upstream in self._upstreams:
            async for path in upstream.walk(directory):
                if path not in seen:
                    seen.add(path)
                    yield path

    @override
    async def get(self, path: Path, /) -> Path:
        for upstream in self._upstreams:
            with suppress(Exception):
                return await upstream.get(path)
        raise UnknownAsset(path, self.directories)


class StaticAssetRepository(AssetRepository):
    """
    Manages static assets.
    """

    def __init__(self, *directories: Path):
        """
        :param directories: Earlier paths have priority over later paths.
        """
        self._directories = directories
        self.__assets: Mapping[Path, Path] | None = None
        self._lock = AsynchronizedLock.new_threadsafe()

    async def _assets(self) -> Mapping[Path, Path]:
        if self.__assets is None:
            async with self._lock:
                self.__assets = await to_thread(self._init_assets)
        return self.__assets

    def _init_assets(self) -> Mapping[Path, Path]:
        return {
            (Path(directory_path) / file_name).relative_to(assets_directory_path): Path(
                directory_path
            )
            / file_name
            for assets_directory_path in reversed(self._directories)
            for directory_path, _, file_names in walk(assets_directory_path)
            for file_name in file_names
        }

    @override
    @property
    def directories(self) -> Sequence[Path]:
        return self._directories

    @override
    async def walk(self, directory: Path | None = None, /) -> AsyncIterable[Path]:
        asset_directory_path_str = str(directory)
        for asset_path in await self._assets():
            if directory is None or str(asset_path).startswith(
                asset_directory_path_str
            ):
                yield asset_path

    @override
    async def get(self, path: Path, /) -> Path:
        try:
            return (await self._assets())[path]
        except KeyError:
            raise UnknownAsset(path, self.directories) from None



"""
Integrate the console and assertion APIs.
"""

import argparse
from collections.abc import Callable

from betty.assertion import Assertion
from betty.exception import HumanFacingException
from betty.locale.localize import Localizer


def assertion_to_argument_type[T](
    assertion: Assertion[str, T], *, localizer: Localizer
) -> Callable[[str], T]:
    """
    Convert an assertion to an argparse argument type.
    """

    def _assertion_to_argument_type(value: str) -> T:
        try:
            return assertion(value)
        except HumanFacingException as error:
            raise argparse.ArgumentTypeError(error.localize(localizer)) from error

    return _assertion_to_argument_type



"""
Presence roles.
"""

from __future__ import annotations

from typing import TYPE_CHECKING, final, override

from betty.definition.human_facing import CountableHumanFacingDefinition
from betty.locale.localizable.gettext import _, ngettext
from betty.plugin import Plugin, PluginDefinition, PluginTypeDefinition
from betty.plugin.discovery.entry_point import EntryPointDiscovery
from betty.plugin.factory import PluginManufacturer
from betty.service.requirement.project import require_project

if TYPE_CHECKING:
    import builtins


class Role(Plugin["RoleDefinition"]):
    """
    A person's role at an event.
    """


@final
@PluginTypeDefinition(
    "role",
    label=_("Role"),
    label_plural=_("Roles"),
    label_countable=ngettext("{count} role", "{count} roles"),
    discovery=[
        EntryPointDiscovery("betty.role"),
        require_project(
            lambda project: (
                configuration.new_plugin()
                for configuration in project.configuration.roles
            )
        ),
    ],
)
class RoleDefinition(CountableHumanFacingDefinition, PluginDefinition[Role]):
    """
    .. plugin_type:: role.
    """


@final
class RoleManufacturer(PluginManufacturer[RoleDefinition, Role]):
    """
    The role manufacturer.
    """

    @override
    @classmethod
    def type(cls) -> builtins.type[RoleDefinition]:
        return RoleDefinition



"""
Boolean data types.
"""

from __future__ import annotations

from typing import TYPE_CHECKING, final

from betty.assertion import assert_bool
from betty.data import DataDefinition
from betty.functools import passthrough
from betty.portable import CallbackPorter

if TYPE_CHECKING:
    from betty.locale.localizable import ResolvableLocalizable


@final
class BoolDefinition(DataDefinition[bool]):
    """
    A boolean data definition.
    """

    def __init__(
        self,
        *,
        label: ResolvableLocalizable,
        description: ResolvableLocalizable | None = None,
    ):
        super().__init__(
            cls=bool,
            label=label,
            description=description,
            porter=CallbackPorter(assert_bool, passthrough),
        )



"""
Provide utilities for concurrent programming.
"""

import asyncio
import threading
import time
from abc import ABC, abstractmethod
from asyncio import sleep
from collections.abc import AsyncIterator, Hashable, MutableMapping
from math import floor
from types import TracebackType
from typing import Self, final, override

from betty.typing import threadsafe

MAX_STRANDS = 64


class Lock(ABC):
    """
    Provide an asynchronous lock.
    """

    async def __aenter__(self):
        await self.acquire()

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        await self.release()

    @abstractmethod
    async def acquire(self, *, wait: bool = True) -> bool:
        """
        Acquire the lock.
        """

    @abstractmethod
    async def release(self) -> None:
        """
        Release the lock.
        """


@final
class AsynchronizedLock(Lock):
    """
    Make a synchronous (blocking) lock asynchronous (non-blocking).
    """

    __slots__ = "_lock"

    def __init__(self, lock: threading.Lock):
        self._lock = lock

    @property
    def lock(self) -> threading.Lock:
        """
        The underlying, synchronous lock.
        """
        return self._lock

    @override
    async def acquire(self, *, wait: bool = True) -> bool:
        while not self._lock.acquire(blocking=False):
            if not wait:
                return False
            # Sleeping for zero seconds does not actually sleep, but gives the event
            # loop a chance to progress other tasks while we wait for another chance
            # to acquire the acquirable.
            await sleep(0)
        return True

    @override
    async def release(self) -> None:
        self._lock.release()

    @classmethod
    def new_threadsafe(cls) -> Self:
        """
        Create a new thread-safe, asynchronous lock.
        """
        return cls(threading.Lock())


@final
@threadsafe
class RateLimiter:
    """
    Rate-limit operations.

    This class implements the `Token Bucket algorithm <https://en.wikipedia.org/wiki/Token_bucket>`_.
    """

    def __init__(self, maximum: int, period: int = 1, /):
        self._lock = AsynchronizedLock.new_threadsafe()
        self._maximum = maximum
        self._period = period
        self._available = maximum
        # A Token Bucket fills as time passes. However, we want callers to be able to start
        # using the limiter immediately, so we 'preload' the first's period's tokens, and
        # set the last added time to the end of the first period. This ensures there is no
        # needless waiting if the number of tokens consumed in total is less than the limit
        # per period.
        self._last_add = time.monotonic() + self._period

    def _add_tokens(self):
        now = time.monotonic()
        elapsed = now - self._last_add
        added = elapsed * self._maximum
        possibly_available = floor(self._available + added)
        if possibly_available > 0:
            self._available = min(possibly_available, self._maximum)
            self._last_add = now

    async def __aenter__(self) -> None:
        await self.wait()

    async def __aexit__(
        self,
        exc_type: type[BaseException] | None,
        exc_val: BaseException | None,
        exc_tb: TracebackType | None,
    ) -> None:
        return

    async def is_available(self) -> bool:
        """
        Whether an operation may be performed (again).
        """
        async with self._lock:
            self._add_tokens()
            return self._available != 0

    async def wait(self) -> None:
        """
        Wait until an operation may be performed (again).
        """
        async with self._lock:
            while self._available < 1:
                self._add_tokens()
                if self._available < 1:
                    await asyncio.sleep(0)
            self._available -= 1


class _Transaction(Lock):
    def __init__(
        self,
        transaction_id: Hashable,
        ledger_lock: Lock,
        ledger: MutableMapping[Hashable, bool],
    ):
        self._transaction_id = transaction_id
        self._ledger_lock = ledger_lock
        self._ledger = ledger

    @override
    async def acquire(self, *, wait: bool = True) -> bool:
        if wait:
            while True:
                async with self._ledger_lock:
                    if self._can_acquire():
                        return self._acquire()
                await sleep(0)
        else:
            async with self._ledger_lock:
                if self._can_acquire():
                    return self._acquire()
                return False

    def _can_acquire(self) -> bool:
        try:
            return not self._ledger[self._transaction_id]
        except KeyError:
            self._ledger[self._transaction_id] = False
            return True

    def _acquire(self) -> bool:
        self._ledger[self._transaction_id] = True
        return True

    @override
    async def release(self) -> None:
        self._ledger[self._transaction_id] = False


@threadsafe
class Ledger:
    """
    Lazily create locks by keeping a ledger.

    The ledger lock is released once a transaction lock is acquired.
    """

    def __init__(self, ledger_lock: Lock):
        self._ledger_lock = ledger_lock
        self._ledger: MutableMapping[Hashable, bool] = {}

    def ledger(self, transaction_id: Hashable) -> Lock:
        """
        Ledger a new lock for the given transaction ID.
        """
        return _Transaction(transaction_id, self._ledger_lock, self._ledger)


async def backoff() -> AsyncIterator[int]:
    """
    Implement `exponential backoff <https://en.wikipedia.org/wiki/Exponential_backoff>`__.

    The returned iterator sleeps after every iteration, increasing the duration with every iteration, up to a limit.

    Usage:

    .. code-block:: python

       async for iteration in backoff():
         if success:
            return  # Or break.
    """
    iterations = 0
    while True:
        yield iterations
        await asyncio.sleep(0.001 * 2 ** min(iterations, 7))
        iterations += 1



from __future__ import annotations  # noqa: D100

from typing import TYPE_CHECKING, Self, final, override

from betty.argparse import assertion_to_argument_type
from betty.assertion import assert_directory_path, assert_none, assert_or
from betty.console.command import Command, CommandDefinition, CommandFunction
from betty.extension import ExtensionDefinition
from betty.locale.localizable.gettext import _
from betty.locale.translation.project import extension as translation_project_extension
from betty.locale.translation.project.extension import (
    assert_extension_has_assets_directory_path,
)
from betty.service.factory import Manufacturable
from betty.service.requirement.app import require_app

if TYPE_CHECKING:
    import argparse
    from pathlib import Path

    from betty.app import App


@final
@CommandDefinition(
    "extension-update-translations",
    label=_("Update all existing translations for an extension"),
)
class ExtensionUpdateTranslations(Manufacturable, Command):
    """
    .. plugin:: command:extension-update-translations.
    """

    def __init__(self, app: App):
        self._app = app

    @override
    @classmethod
    @require_app
    async def new(cls, app: App, /) -> Self:
        return cls(app)

    @override
    async def configure(self, parser: argparse.ArgumentParser) -> CommandFunction:
        extensions = await self._app.plugins.plugins(ExtensionDefinition)
        localizer = await self._app.localizer

        parser.add_argument(
            "extension",
            type=assertion_to_argument_type(
                lambda extension_id: assert_extension_has_assets_directory_path(
                    extensions[extension_id]
                ),
                localizer=localizer,
            ),
        )
        parser.add_argument(
            "source",
            type=assertion_to_argument_type(
                assert_or(assert_none, assert_directory_path()), localizer=localizer
            ),
        )
        parser.add_argument(
            "--exclude",
            action="append",
            type=assertion_to_argument_type(
                assert_directory_path(), localizer=localizer
            ),
        )
        return self._command_function

    async def _command_function(
        self, extension: ExtensionDefinition, source: Path, exclude: tuple[Path] | None
    ) -> None:
        await translation_project_extension.update_extension_translations(
            extension, source, None if exclude is None else set(exclude)
        )



"""
An API to interact with Betty's user.
"""

from __future__ import annotations

from abc import ABC, abstractmethod
from enum import IntEnum
from typing import TYPE_CHECKING, overload

from betty.locale.localize import DEFAULT_LOCALIZER
from betty.typing import Void, VoidType

if TYPE_CHECKING:
    import logging
    from contextlib import AbstractAsyncContextManager

    from betty.assertion import Assertion
    from betty.locale.localizable import ResolvableLocalizable
    from betty.locale.localize import Localizer
    from betty.progress import Progress


class Verbosity(IntEnum):
    """
    User interaction verbosity.
    """

    QUIET = -1
    """
    Inform users of errors, but do not show any other output.
    """

    DEFAULT = 0
    """
    Like QUIET, and show warning and information summary messages.
    """

    VERBOSE = 1
    """
    Like DEFAULT, and show information details messages,
    """

    MORE_VERBOSE = 2
    """
    Like VERBOSE, and show debug messages.
    """

    MOST_VERBOSE = 3
    """
    Like MORE_VERBOSE, and show all log messages.
    """


class UserError(Exception):
    """
    A user session error.
    """


class UserTimeoutError(UserError):
    """
    The user did not respond within the given time, or at all.
    """


class User(ABC):
    """
    A user session.
    """

    localizer: Localizer = DEFAULT_LOCALIZER

    @property
    @abstractmethod
    def verbosity(self) -> Verbosity:
        """
        The current verbosity.
        """

    @abstractmethod
    async def set_verbosity(self, verbosity: Verbosity, /) -> None:
        """
        Set the new verbosity.
        """

    @abstractmethod
    async def message_exception(self) -> None:
        """
        Send an error message about an exception to the user.

        An error indicates that something went wrong, and Betty was unable to perform an expected action.

        These messages are always shown to the user.
        """

    @abstractmethod
    async def message_error(self, message: ResolvableLocalizable, /) -> None:
        """
        Send an error message to the user.

        An error indicates that something went wrong, and Betty was unable to perform an expected action.

        These messages are always shown to the user.
        """

    @abstractmethod
    async def message_warning(self, message: ResolvableLocalizable, /) -> None:
        """
        Send a warning message to the user.

        A warning indicates that something went wrong, but that Betty was able to recover and perform an expected
        action, but perhaps in a slightly different way.

        These messages are shown to users for :py:attr:`betty.user.Verbosity.DEFAULT` and up.
        """

    @abstractmethod
    async def message_information(self, message: ResolvableLocalizable, /) -> None:
        """
        Send a summarized informative message to the user.

        An informative message tells the user that something happened successfully, e.g. the starting or finishing of a
        task.

        These messages are shown to users for :py:attr:`betty.user.Verbosity.DEFAULT` and up.
        """

    @abstractmethod
    async def message_information_details(
        self, message: ResolvableLocalizable, /
    ) -> None:
        """
        Send a detailed informative message to the user.

        An informative message tells the user that something happened successfully, e.g. the starting or finishing of a
        task.

        These messages are shown to users for :py:attr:`betty.user.Verbosity.VERBOSE` and up.
        """

    @abstractmethod
    async def message_debug(self, message: ResolvableLocalizable, /) -> None:
        """
        Send a debugging message to the user.

        These messages are shown to users for :py:attr:`betty.user.Verbosity.MORE_VERBOSE` and up.
        """

    @abstractmethod
    async def message_log(self, message: logging.LogRecord, /) -> None:
        """
        Send a log message to the user.

        These messages are shown to users for :py:attr:`betty.user.Verbosity.MOST_VERBOSE` and up.
        """

    @abstractmethod
    def message_progress(
        self, message: ResolvableLocalizable, /
    ) -> AbstractAsyncContextManager[Progress]:
        """
        Send information about a progressing activity to the user.

        These messages are shown to users for :py:attr:`betty.user.Verbosity.DEFAULT` and up.
        """

    @abstractmethod
    async def ask_confirmation(
        self, statement: ResolvableLocalizable, *, default: bool = False
    ) -> bool:
        """
        Ask the user to confirm a statement.

        :raises: betty.user.UserTimeoutError
        """

    @overload
    async def ask_input(
        self,
        question: ResolvableLocalizable,
        *,
        default: str | VoidType = Void,
    ) -> str:
        pass

    @overload
    async def ask_input[T](
        self,
        question: ResolvableLocalizable,
        *,
        assertion: Assertion[str, T],
        default: str | VoidType = Void,
    ) -> T:
        pass

    @abstractmethod
    async def ask_input[T](
        self,
        question: ResolvableLocalizable,
        *,
        assertion: Assertion[str, T] | None = None,
        default: str | T | VoidType = Void,
    ) -> str | T:
        """
        Ask the user to input text.

        :raises: betty.user.UserTimeoutError
        """



"""
Service providers.
"""

from __future__ import annotations

from abc import abstractmethod
from collections.abc import Awaitable, Callable
from functools import update_wrapper
from inspect import iscoroutinefunction
from typing import TYPE_CHECKING, Any, Protocol, Self, cast, overload, override

from betty.concurrent import AsynchronizedLock, Lock
from betty.life_cycle.manage import ManagedLifeCycle
from betty.service import ServiceError
from betty.typing import Void, VoidType

if TYPE_CHECKING:
    from types import FunctionType

    from ty_extensions import Intersection


type ServiceFactory[ServiceProviderT, ServiceT] = Callable[[ServiceProviderT], ServiceT]


class _ServiceDecorator[ServiceProviderT, ServiceT](Protocol):
    @overload
    def __call__(
        self, factory: Callable[[ServiceProviderT], ServiceT], /
    ) -> _SynchronousServiceManager[ServiceProviderT, ServiceT]:
        pass

    @overload
    def __call__(
        self, factory: Callable[[ServiceProviderT], Awaitable[ServiceT]], /
    ) -> _AsynchronousServiceManager[ServiceProviderT, ServiceT]:
        pass


@overload
def service[ServiceProviderT, ServiceT](
    factory: Callable[[ServiceProviderT], Awaitable[ServiceT]], /
) -> _AsynchronousServiceManager[ServiceProviderT, ServiceT]:
    pass


@overload
def service[ServiceProviderT, ServiceT](
    factory: Callable[[ServiceProviderT], ServiceT], /
) -> _SynchronousServiceManager[ServiceProviderT, ServiceT]:
    pass


@overload
def service(factory: None = None, /) -> _ServiceDecorator:
    pass


def service(factory):
    """
    Decorate a service factory method.

    The factory method is replaced with a :py:class:`service manager <betty.service.provider.ServiceManager>` which
    handles lazy service instantiation, caching, and multiprocessing support.

    The decorated factory method should return a new service instance.
    """

    def _service[ServiceProviderT, ServiceT](
        factory: Callable[[ServiceProviderT], ServiceT], /
    ) -> ServiceManager[ServiceProviderT, ServiceT, Any]:
        if iscoroutinefunction(factory):
            return _AsynchronousServiceManager(
                factory,  # ty:ignore[invalid-argument-type]
            )  # ty:ignore[invalid-return-type]
        return _SynchronousServiceManager(
            factory,  # ty:ignore[invalid-argument-type]
        )

    if factory is None:
        return _service
    return _service(factory)


class ServiceManager[ServiceProviderT, ServiceGetT, ServiceT]:
    """
    Manages a single service for a service provider.
    """

    def __init__(
        self,
        factory: Intersection[
            ServiceFactory[ServiceProviderT, ServiceGetT], FunctionType
        ],
        /,
    ):
        update_wrapper(
            self,  # ty:ignore[invalid-argument-type]
            factory,
        )
        self._factory = factory
        self._service_name: str = factory.__name__
        self._service_attr_name = f"_{self._service_name}"
        self._service_override_attr_name = f"{self._service_attr_name}_override"
        self._factory_override_attr_name = f"{self._service_attr_name}_factory_override"

    @property
    def name(self) -> str:
        """
        The service name.
        """
        return self._service_name

    @overload
    def __get__(self, instance: None, owner: type[ServiceProviderT]) -> Self:
        pass

    @overload
    def __get__(
        self, instance: ServiceProviderT, owner: type[ServiceProviderT]
    ) -> ServiceGetT:
        pass

    def __get__(
        self, instance: ServiceProviderT | None, owner: type[ServiceProviderT]
    ) -> ServiceGetT | Self:
        if instance is None:
            return self

        return self.get(instance)

    def get(self, instance: ServiceProviderT, /) -> ServiceGetT:
        """
        Get the service from an instance.
        """
        if isinstance(instance, ManagedLifeCycle):
            instance.assert_alive()

        return self._get(instance)

    @abstractmethod
    def _get(self, instance: ServiceProviderT, /) -> ServiceGetT:
        pass

    def _get_attr(self, instance: ServiceProviderT, /) -> ServiceT | VoidType:
        return getattr(instance, self._service_attr_name, Void)

    def _get_factory(
        self, instance: ServiceProviderT, /
    ) -> ServiceFactory[ServiceProviderT, ServiceGetT]:
        factory = cast(
            "ServiceFactory[ServiceProviderT, ServiceGetT] | None",
            getattr(instance, self._factory_override_attr_name, None),
        )
        if factory is not None:
            return factory
        return self._factory

    def _assert_not_initialized(self, instance: ServiceProviderT, /):
        if self._get_attr(instance) is not Void:
            raise ServiceInitializedError(
                f"{instance}.{self._service_name} was initialized already."
            )

    def override(self, instance: ServiceProviderT, service: ServiceT, /) -> None:
        """
        Override the service for the given instance.

        Calling this will prevent any existing factory from being called.

        This MUST only be called from ``instance.__init__()``.
        """
        self._assert_not_initialized(instance)
        setattr(instance, self._service_attr_name, service)
        setattr(instance, self._service_override_attr_name, True)

    def override_factory(
        self,
        instance: ServiceProviderT,
        factory: ServiceFactory[ServiceProviderT, ServiceGetT],
        /,
    ) -> None:
        """
        Override the default service factory for the given instance.

        This MUST only be called from ``instance.__init__()``. It will override the existing service factory method
        defined on the instance.
        """
        self._assert_not_initialized(instance)
        setattr(instance, self._factory_override_attr_name, factory)


class _AsynchronousServiceManager[ServiceProviderT, ServiceT](
    ServiceManager[ServiceProviderT, Awaitable[ServiceT], ServiceT],
):
    def _lock(self, instance: ServiceProviderT, /) -> Lock:
        lock_attr_name = f"_{self._service_attr_name}_lock"
        try:
            return cast(Lock, getattr(instance, lock_attr_name))
        except AttributeError:
            lock = AsynchronizedLock.new_threadsafe()
            setattr(instance, lock_attr_name, lock)
            return lock

    @override
    async def _get(self, instance: ServiceProviderT, /) -> ServiceT:
        async with self._lock(instance):
            service = self._get_attr(instance)

            if service is not Void:
                return service

            new_service = await self._get_factory(instance)(instance)
            setattr(instance, self._service_attr_name, new_service)
            return new_service


class _SynchronousServiceManager[ServiceProviderT, ServiceT](
    ServiceManager[ServiceProviderT, ServiceT, ServiceT]
):
    @override
    def _get(self, instance: ServiceProviderT, /) -> ServiceT:
        service = self._get_attr(instance)
        if service is not Void:
            return service

        new_service = self._get_factory(instance)(instance)
        setattr(instance, self._service_attr_name, new_service)
        return new_service


class ServiceInitializedError(ServiceError):
    """
    A service was unexpectedly initialized already.
    """



"""
Jobs for the Trees extension.
"""

from __future__ import annotations

import json
from asyncio import gather
from typing import TYPE_CHECKING, override

import aiofiles
from aiofiles.os import makedirs

from betty.ancestry.person import Person
from betty.job import Job
from betty.media_type.media_types import HTML

if TYPE_CHECKING:
    from babel import Locale

    from betty.job.scheduler import Scheduler
    from betty.project import Project


class _GeneratePeopleJson(Job):
    def __init__(self, *, project: Project):
        super().__init__("trees:generate-people-json")
        self._project = project

    @override
    async def do(self, scheduler: Scheduler, /) -> None:
        await gather(
            *(
                self._generate_people_json_for_locale(scheduler, locale)
                for locale in self._project.configuration.locales.keys()  # noqa: SIM118
            )
        )

    async def _generate_people_json_for_locale(
        self, scheduler: Scheduler, locale: Locale
    ) -> None:
        url_generator = await self._project.url_generator
        localizers = await self._project.localizers
        localizer = localizers.get(locale)
        private_label = localizer._("private")
        people = {
            person.id: {
                "id": person.id,
                "label": person.label.localize(localizer)
                if person.public
                else private_label,
                "url": url_generator.generate(person, media_type=HTML),
                "parentIds": [parent.id for parent in person.parents],
                "childIds": [child.id for child in person.children],
                "private": person.private,
            }
            for person in self._project.ancestry[Person]
        }
        people_json = json.dumps(people)
        await makedirs(self._project.localize_www_directory(locale), exist_ok=True)
        async with aiofiles.open(
            self._project.localize_www_directory(locale) / "people.json",
            mode="w",
        ) as f:
            await f.write(people_json)



"""
Collection data types.
"""

from __future__ import annotations

from collections.abc import Callable, Collection
from typing import TYPE_CHECKING, Any

from betty.data import DataDefinition
from betty.data.aggregate import AggregateDefinition
from betty.data.indicator.selector import Element

if TYPE_CHECKING:
    from betty.data import Data
    from betty.locale.localizable import ResolvableLocalizable
    from betty.portable import Porter


class CollectionDefinition[CollectionT: Collection, ElementT: Element[Any]](
    AggregateDefinition[CollectionT, ElementT]
):
    """
    A homogenous collection data definition.
    """

    def __init__(
        self,
        /,
        cls: type[CollectionT] | None = None,
        *,
        item: DataDefinition | type[Data],
        label: ResolvableLocalizable,
        description: ResolvableLocalizable | None = None,
        porter: Porter[CollectionT] | None = None,
        factory: Callable[[], CollectionT] | None = None,
    ):
        super().__init__(cls=cls, label=label, description=description, porter=porter)
        self._item = item if isinstance(item, DataDefinition) else item.data()
        self._factory = factory

    @property
    def item(self) -> DataDefinition:
        """
        The definition of the items contained by this collection.
        """
        return self._item

    def new(self) -> CollectionT:
        """
        Create a new collection.
        """
        return (self.cls if not self._factory else self._factory)()



"""
Presence roles.
"""

from typing import final

from betty.classtools import Singleton
from betty.locale.localizable.gettext import _, ngettext
from betty.role import Role, RoleDefinition


@final
@RoleDefinition(
    "attendee",
    label=_("Attendee"),
    label_plural=_("Attendees"),
    label_countable=ngettext("{count} attendee", "{count} attendees"),
)
class Attendee(Role):
    """
    .. plugin:: role:attendee.
    """


@final
@RoleDefinition(
    "beneficiary",
    label=_("Beneficiary"),
    label_plural=_("Beneficiaries"),
    label_countable=ngettext("{count} beneficiary", "{count} beneficiaries"),
)
class Beneficiary(Role):
    """
    .. plugin:: role:beneficiary.

    Someone was a `benificiary <https://en.wikipedia.org/wiki/Beneficiary>`_ in the event, such as a :py:class:`betty.event_type.event_types.Will`.
    """


@final
@RoleDefinition(
    "celebrant",
    label=_("Celebrant"),
    label_plural=_("Celebrants"),
    label_countable=ngettext("{count} celebrant", "{count} celebrants"),
)
class Celebrant(Role):
    """
    .. plugin:: role:celebrant.

    Someone was the `celebrant <https://en.wikipedia.org/wiki/Officiant>`_ at the event.

    This includes but is not limited to:

    - civil servant
    - religious leader
    - civilian
    """


@final
@RoleDefinition(
    "informant",
    label=_("Informant"),
    label_plural=_("Informants"),
    label_countable=ngettext("{count} informant", "{count} informants"),
    description=_("Someone reported the event with a record-keeping institution."),
)
class Informant(Role):
    """
    .. plugin:: role:informant.
    """


@final
@RoleDefinition(
    "organizer",
    label=_("Organizer"),
    label_plural=_("Organizers"),
    label_countable=ngettext("{count} organizer", "{count} organizers"),
)
class Organizer(Role):
    """
    .. plugin:: role:organizer.
    """


@final
@RoleDefinition(
    "speaker",
    label=_("Speaker"),
    label_plural=_("Speakers"),
    label_countable=ngettext("{count} speaker", "{count} speakers"),
    description=_("Someone performed public speaking at the event."),
)
class Speaker(Role):
    """
    .. plugin:: role:speaker.
    """


@final
@RoleDefinition(
    "subject",
    label=_("Subject"),
    label_plural=_("Subjects"),
    label_countable=ngettext("{count} subjects", "{count} subjects"),
)
class Subject(Role):
    """
    .. plugin:: role:subject.

    The meaning of this role depends on the event type. For example, for
    :py:class:`betty.event_type.event_types.Marriage`, the subjects are the people who got married. For
    :py:class:`betty.event_type.event_types.Death` it is the person who died.
    """


@final
@RoleDefinition(
    "unknown",
    label=_("Unknown"),
    label_plural=_("Unknowns"),
    label_countable=ngettext("{count} unknown", "{count} unknowns"),
)
class Unknown(Role, Singleton):
    """
    .. plugin:: role:unknown.
    """


@final
@RoleDefinition(
    "witness",
    label=_("Witness"),
    label_plural=_("Witnesses"),
    label_countable=ngettext("{count} witness", "{count} witnesses"),
    description=_("A formal witness to an event."),
)
class Witness(Role):
    """
    .. plugin:: role:witness.
    """



"""
Integrate Betty with `Gramps <https://gramps-project.org>`_.
"""

from __future__ import annotations

from typing import TYPE_CHECKING, Self, final, override

from betty.extension import Extension, ExtensionDefinition
from betty.extension.gramps.data import FamilyTree, GrampsConfiguration
from betty.extension.gramps.jobs import LoadAncestry
from betty.gramps.loader import GrampsLoader
from betty.locale.localizable.gettext import _
from betty.project.load import Loader
from betty.service.factory import DataManufacturable, Manufacturable
from betty.service.requirement.project import require_project

if TYPE_CHECKING:
    from collections.abc import Iterable
    from pathlib import Path

    from betty.ancestry import Ancestry
    from betty.job.scheduler import Scheduler
    from betty.project import Project
    from betty.service.level import ServiceLevel
    from betty.user import User


@final
@ExtensionDefinition(
    "gramps",
    label="Gramps",
    description=_("Load Gramps family trees."),
)
class Gramps(
    Loader, DataManufacturable[GrampsConfiguration], Manufacturable, Extension
):
    """
    .. plugin:: extension:gramps.

    Attributes
    ----------
    Gramps allows arbitrary attributes to be added to some of its data types. Betty can parse these to load additional
    information. Each of Betty's Gramps attributes follows the same structure: ``betty:...`` (to load the attribute for any
    Betty project) or ``betty-MyProject:..`` (to load an attribute for the Betty project with machine name ``MyProject``),
    where ``...`` is the name that identifies the attribute's meaning. For the 'privacy` attribute, the Gramps attribute's full
    name would be ``betty:privacy`` or ``betty-MyProject:privacy``.

    Privacy
    ^^^^^^^

    Gramps has limited built-in support for people's privacy. To fully control privacy for people, as well as events, files,
    sources, and citations, add a ``betty:privacy`` attribute to any of these types, with a value of ``private`` to explicitly
    declare the data always private or ``public`` to declare the data always public. Any other value will leave the privacy
    undecided, as well as person records marked public using Gramps' built-in privacy selector. In such cases, the
    ``privatizer`` extension may decide if the data is public or private.

    Gender
    ^^^^^^
    To set a person's gender to a gender that is available in Betty, but not in Gramps, add a ``betty:gender`` attribute,
    whose value is the ID of the :py:class:`gender plugin <betty.gender.GenderDefinition>` you want to use.

    Event names
    ^^^^^^^^^^^
    Event names can be set using ``betty:name``. Values are :ref:`static translations <gramps-attributes-static-translations>`.

    Links
    ^^^^^

    Gramps has limited built-in support to add links to entities. For those Gramps entities that support attributes,
    you may add links using those:

    .. list-table:: Link attributes
       :header-rows: 1

       * - Name
         - Required/optional
         - Description
       * - ``betty:link-LINKNAME:url``
         - **required**
         - The URL the link targets. This may contain :ref:`static translations <gramps-attributes-static-translations>`.
       * - ``betty:link-LINKNAME:description``
         - optional
         - A human-friendly longer link description. This may contain :ref:`static translations <gramps-attributes-static-translations>`.
       * - ``betty:link-LINKNAME:label``
         - optional
         - A human-friendly short link label. This may contain :ref:`static translations <gramps-attributes-static-translations>`.
       * - ``betty:link-LINKNAME:media_type``
         - optional
         - An `IANA media type <https://www.iana.org/assignments/media-types/media-types.xhtml>`_.
       * - ``betty:link-LINKNAME:relationship``
         - optional
         - An `IANA link relationship <https://www.iana.org/assignments/link-relations/link-relations.xhtml>`_.

    Where ``LINKNAME`` may be any value of your choosing, but must be unique per link. For example, where ``LINKNAME`` is ``cheese``:

    .. list-table::

       * - ``betty:link-cheese:url``
         - ``https://en.wikipedia.org/wiki/Cheese``
       * - ``betty:link-cheese:label``
         - ``Learn about cheese``
       * - ``betty:link-cheese:description``
         - ``Read the Wikipedia article about cheese``

    .. _gramps-attributes-static-translations:

    Static translations
    ^^^^^^^^^^^^^^^^^^^
    Static translations are not attributes on their own per se, but they are used by other attributes, such as links.

    If another attribute defines itself as containing static translations, that means you may add multiple variants of the
    attribute, each with a translation for a different locale.

    For example, given a translatable attribute called ``betty:my-text``, you may add an actual attribute ``betty:my-text``
    with any translation for a locale which Betty will consider *undetermined*. You may also add any number of
    ``betty:my-text:LOCALE`` attributes, where ``LOCALE`` is an `IETF BCP 47 <https://tools.ietf.org/html/bcp47>`_ language
    tag, and the value is the translation for that specific locale/language.

    Dates
    -----

    For unknown date parts, set those to all zeroes and Betty will ignore them. For instance, ``0000-12-31`` will be parsed as
    "December 31", and ``1970-01-00`` as "January, 1970".

    Event types
    -----------

    Betty supports the following Gramps event types without any additional configuration:

    .. list-table:: Event types
       :align: left
       :header-rows: 1

       * - Gramps event type
         - Betty event type
       * - ``Adopted``
         - ``adoption``
       * - ``Adult Christening``
         - ``baptism``
       * - ``Baptism``
         - ``baptism``
       * - ``Bar Mitzvah``
         - ``bar-mitzvah``
       * - ``Bat Mitzvah``
         - ``bat-mitzvah``
       * - ``Birth``
         - ``birth``
       * - ``Burial``
         - ``burial``
       * - ``Christening``
         - ``baptism``
       * - ``Confirmation``
         - ``confirmation``
       * - ``Cremation``
         - ``cremation``
       * - ``Death``
         - ``death``
       * - ``Divorce``
         - ``divorce``
       * - ``Divorce Filing``
         - ``divorce-announcement``
       * - ``Emigration``
         - ``emigration``
       * - ``Engagement``
         - ``engagement``
       * - ``Immigration``
         - ``immigration``
       * - ``Marriage``
         - ``marriage``
       * - ``Marriage Banns``
         - ``marriage-announcement``
       * - ``Occupation``
         - ``occupation``
       * - ``Residence``
         - ``residence``
       * - ``Retirement``
         - ``retirement``
       * - ``Will``
         - ``will``

    Genders
    -------

    Betty maps Gramps genders as follows:

    .. list-table:: Genders
       :align: left
       :header-rows: 1

       * - Gramps gender
         - Betty gender
       * - ``F``
         - ``woman``
       * - ``M``
         - ``man``
       * - ``U``
         - ``unknown``
       * - ``X``
         - ``non-binary``

    Place types
    -----------

    Betty supports the following Gramps place types without any additional configuration:

    .. list-table:: Place types
       :align: left
       :header-rows: 1

       * - Gramps place type
         - Betty place type
       * - ``Borough``
         - ``borough``
       * - ``Building``
         - ``building``
       * - ``City``
         - ``city``
       * - ``Country``
         - ``country``
       * - ``County``
         - ``county``
       * - ``Department``
         - ``department``
       * - ``District``
         - ``district``
       * - ``Farm``
         - ``farm``
       * - ``Hamlet``
         - ``hamlet``
       * - ``Locality``
         - ``locality``
       * - ``Municipality``
         - ``municipality``
       * - ``Neighborhood``
         - ``neighborhood``
       * - ``Number``
         - ``number``
       * - ``Parish``
         - ``parish``
       * - ``Province``
         - ``province``
       * - ``Region``
         - ``region``
       * - ``State``
         - ``state``
       * - ``Street``
         - ``street``
       * - ``Town``
         - ``town``
       * - ``Unknown``
         - ``Unknown``
       * - ``Village``
         - ``village``

    Roles
    -----

    Betty supports the following Gramps roles without any additional configuration:

    .. list-table:: Roles
       :align: left
       :header-rows: 1

       * - Gramps role
         - Betty role
       * - ``Aide``
         - ``attendee``
       * - ``Bride``
         - ``subject``
       * - ``Celebrant``
         - ``celebrant``
       * - ``Clergy``
         - ``celebrant``
       * - ``Family``
         - ``subject``
       * - ``Groom``
         - ``subject``
       * - ``Informant``
         - ``informant``
       * - ``Primary``
         - ``subject``
       * - ``Unknown``
         - ``unknown``
       * - ``Witness``
         - ``witness``

    Order & priority
    ----------------

    The order of lists of data, or the priority of individual bits of data, can be automatically determined by Betty in
    multiple different ways, such as by matching dates, or locales. When not enough details are available, or in case of
    ambiguity, the original order is preserved. If only a single item must be retrieved from the list, this will be the
    first item, optionally after sorting.

    For example, if a place has multiple names (which may be historical or translations), Betty may try to
    filter names by the given locale and date, and then indiscriminately pick the first one of the remaining names to
    display as the canonical name.

    Tips:

    - If you want one item to have priority over another, it should come before the other in a list (e.g. be higher up).
    - Items with more specific or complete data, such as locales or dates, should come before items with less specific or
      complete data. However, items without dates at all are considered current and not historical.
    - Unofficial names or nicknames, should generally be put at the end of lists.

    """

    def __init__(
        self,
        *,
        ancestry: Ancestry,
        services: ServiceLevel,
        user: User,
        attribute_prefix_key: str | None = None,
        executable: Path | None = None,
        family_trees: Iterable[FamilyTree] | None = None,
    ):
        super().__init__()
        self._services = services
        self._ancestry = ancestry
        self._attribute_prefix_key = attribute_prefix_key
        self._executable = executable
        self._family_trees = () if family_trees is None else list(family_trees)
        self._user = user

    @override
    @classmethod
    def new_data_cls(cls) -> type[GrampsConfiguration]:
        return GrampsConfiguration

    @override
    @classmethod
    @require_project
    async def new(
        cls, project: Project, data: GrampsConfiguration | None = None, /
    ) -> Self:
        return cls(
            ancestry=project.ancestry,
            executable=None if data is None else data.executable,
            family_trees=None if data is None else data.family_trees,
            services=project,
            user=project.app.user,
        )

    @override
    async def load(self, scheduler: Scheduler) -> None:
        for family_tree in self._family_trees:
            await scheduler.add(
                LoadAncestry(
                    loader=GrampsLoader(
                        self._ancestry,
                        services=self._services,
                        attribute_prefix_key=self._attribute_prefix_key,
                        user=self._user,
                        event_type_mapping=family_tree.event_types,
                        place_type_mapping=family_tree.place_types,
                        role_mapping=family_tree.roles,
                        executable=self._executable,
                    ),
                    source=family_tree.source,
                )
            )



"""
Data selectors.
"""

from __future__ import annotations

from abc import ABC, abstractmethod
from contextlib import contextmanager
from typing import TYPE_CHECKING, Any, final, override

from betty.data.indicator import Indicator

if TYPE_CHECKING:
    from collections.abc import Iterator, MutableSequence, Sequence

    from betty.assertion import Assertion


@final
class SelectorError(ValueError):
    """
    Raise when a selector cannot access its element on the given data.
    """

    def __init__(self, selector: Selector, /):
        super().__init__(f"Cannot access {selector.format()}")


class Selector(Indicator, ABC):
    """
    Indicate and interact with aggregate data.
    """

    @contextmanager
    def _catch(self) -> Iterator[None]:
        try:
            yield
        except Exception as error:
            raise SelectorError(self) from error

    @final
    def get[T](self, data: Any, assertion: Assertion[Any, T] | None = None, /) -> T:
        """
        Get the value for this selector.

        :raises SelectorError: raised if the selected element cannot be accessed.
        """
        with self._catch():
            data = self._get(data)
        return assertion(data) if assertion else data

    @abstractmethod
    def _get(self, data: Any, /) -> Any:
        pass

    @final
    def set(self, data: Any, value: Any, /) -> None:
        """
        Set the value for this selector.

        :raises SelectorError: raised if the selected element cannot be accessed.
        """
        with self._catch():
            self._set(data, value)

    @abstractmethod
    def _set(self, data: Any, value: Any, /) -> None:
        pass

    @final
    def delete(self, data: Any, /) -> None:
        """
        Delete the element for this selector.

        :raises SelectorError: raised if the selected element cannot be accessed.
        """
        with self._catch():
            self._delete(data)

    @abstractmethod
    def _delete(self, data: Any, /) -> None:
        pass


@final
class Selectors(Selector):
    """
    Combine multiple selectors into one.
    """

    def __init__(self, *selectors: Selector):
        self._selectors = list(selectors)

    @override
    def format(self) -> str:
        return "".join(["data", *[selector.format() for selector in self._selectors]])

    @classmethod
    def reduce(cls, *indicators: Indicator) -> Sequence[Indicator]:
        """
        Reduce all consecutive instances of py:class:`betty.data.indicator.selector.Selector` to a single instance of this class.

        All other indicators are kept verbatim.
        """
        reduced_indicators: MutableSequence[Indicator] = []
        for indicator in indicators:
            if isinstance(indicator, Selector):
                try:
                    last_indicator = reduced_indicators[-1]
                except IndexError:
                    pass
                else:
                    if isinstance(last_indicator, Selectors):
                        last_indicator._selectors.append(indicator)
                        continue
                reduced_indicators.append(Selectors(indicator))
            else:
                reduced_indicators.append(indicator)
        return reduced_indicators

    @override
    def _get(self, data: Any, /) -> Any:
        for selector in self._selectors:
            data = selector.get(data)
        return data

    @override
    def _set(self, data: Any, value: Any, /) -> None:
        for selector in self._selectors[:-1]:
            data = selector.get(data)
        self._selectors[-1].set(data, value)

    @override
    def _delete(self, data: Any, /) -> None:
        for selector in self._selectors[:-1]:
            data = selector.get(data)
        self._selectors[-1].delete(data)


class Element[ElementT](Selector):
    """
    An aggregate element selector.
    """

    def __init__(self, element: ElementT, /):
        self._element = element

    def __eq__(self, other: Any) -> bool:
        if type(other) is not type(self):
            return NotImplemented
        return self._element == other._element

    @property
    def element(self) -> ElementT:
        """
        The element.
        """
        return self._element


@final
class Attr(Element[str]):
    """
    An attribute selector.
    """

    @override
    def format(self) -> str:
        return f".{self.element}"

    @override
    def _get(self, data: Any, /) -> Any:
        return getattr(data, self.element)

    @override
    def _set(self, data: Any, value: Any, /) -> None:
        setattr(data, self.element, value)

    @override
    def _delete(self, data: Any, /) -> None:
        delattr(data, self.element)


@final
class Index(Element[int]):
    """
    A sequence item selector.
    """

    @override
    def format(self) -> str:
        return f"[{self.element}]"

    @override
    def _get(self, data: Any, /) -> Any:
        return data[self.element]

    @override
    def _set(self, data: Any, value: Any, /) -> None:
        data[self.element] = value

    @override
    def _delete(self, data: Any, /) -> None:
        del data[self.element]


@final
class Key(Element[str]):
    """
    A mapping key selector.
    """

    @override
    def format(self) -> str:
        return f'["{self.element}"]'

    @override
    def _get(self, data: Any, /) -> Any:
        return data[self.element]

    @override
    def _set(self, data: Any, value: Any, /) -> None:
        data[self.element] = value

    @override
    def _delete(self, data: Any, /) -> None:
        del data[self.element]



"""
Test utilities for :py:mod:`betty.json.schema`.
"""

from collections.abc import MutableMapping, Sequence

import pytest
from jsonschema.exceptions import ValidationError

from betty.json.schema import JsonSchemaSchema, Schema, String
from betty.portable import PortableData

DUMMY_SCHEMAS: Sequence[
    tuple[Schema, Sequence[PortableData], Sequence[PortableData]]
] = (
    (
        String(),
        [
            "",
            "Hello, world!",
        ],
        [True, False, None, 123, [], {}],
    ),
    (
        String(def_name="myFirstSchema"),
        ["", "Hello, world!"],
        [True, False, None, 123, [], {}],
    ),
)


type SchemaTestBaseSut = tuple[Schema, Sequence[PortableData], Sequence[PortableData]]


class SchemaTestBase:
    """
    A base class for testing :py:class:`betty.json.schema.Schema` implementations.
    """

    @pytest.fixture
    def sut_data(self) -> SchemaTestBaseSut:
        """
        Provide the system(s) under test.
        """
        raise NotImplementedError

    async def test_def_name(self, sut_data: SchemaTestBaseSut) -> None:
        """
        Tests :py:attr:`betty.json.schema.Schema.def_name` implementations.
        """
        sut, _, __ = sut_data
        assert sut.def_name is None or len(sut.def_name)

    async def test_schema(self, sut_data: SchemaTestBaseSut) -> None:
        """
        Tests :py:attr:`betty.json.schema.Schema.schema` implementations.
        """
        sut, _, __ = sut_data
        assert isinstance(sut.schema, MutableMapping)
        JsonSchemaSchema().validate(sut.schema)

    async def test_defs(self, sut_data: SchemaTestBaseSut) -> None:
        """
        Tests :py:attr:`betty.json.schema.Schema.defs` implementations.
        """
        sut, _, __ = sut_data
        assert isinstance(sut.defs, MutableMapping)

    async def test_embed(self, sut_data: SchemaTestBaseSut) -> None:
        """
        Tests :py:meth:`betty.json.schema.Schema.embed` implementations.
        """
        sut, _, __ = sut_data
        into = Schema()
        assert isinstance(sut.embed(into), MutableMapping)

    async def test_validate_should_validate(self, sut_data: SchemaTestBaseSut) -> None:
        """
        Tests :py:meth:`betty.json.schema.Schema.validate` implementations.
        """
        sut, valid_datas, _invalid_datas = sut_data
        for valid_data in valid_datas:
            sut.validate(valid_data)

    async def test_validate_should_invalidate(
        self, sut_data: SchemaTestBaseSut
    ) -> None:
        """
        Tests :py:meth:`betty.json.schema.Schema.validate` implementations.
        """
        sut, _valid_datas, invalid_datas = sut_data
        for invalid_data in invalid_datas:
            with pytest.raises(ValidationError):
                sut.validate(invalid_data)



"""
Linked data for the localizable API.
"""

from collections.abc import Iterable

from betty.locale import to_language_tag
from betty.locale.localizable import Localizable
from betty.locale.localizable.static import StaticTranslations
from betty.locale.localize import Localizer
from betty.portable import PortableMapping


def dump_linked_data(
    localizable: Localizable, *, localizers: Iterable[Localizer]
) -> PortableMapping:
    """
    Dump a :py:class:`betty.locale.localizable.Localizable` to `JSON-LD <https://json-ld.org/>`_.
    """
    return {
        to_language_tag(locale): translation
        for locale, translation in StaticTranslations.resolve(
            localizable, localizers
        ).translations.items()
    }



"""
Test utilities for :py:mod:`betty.copyright_notice`.
"""

from __future__ import annotations

import pytest

from betty.copyright_notice import CopyrightNotice
from betty.locale.localize import DEFAULT_LOCALIZER


class CopyrightNoticeTestBase[CopyrightNoticeT: CopyrightNotice]:
    """
    A base class for testing :py:class:`betty.copyright_notice.CopyrightNotice` implementations.
    """

    @pytest.fixture
    def sut(self) -> type[CopyrightNoticeT]:
        """
        Provide the system(s) under test.
        """
        raise NotImplementedError

    def test_summary(self, sut: CopyrightNoticeT) -> None:
        """
        Tests :py:meth:`betty.copyright_notice.CopyrightNotice.summary` implementations.
        """
        assert sut.summary.localize(DEFAULT_LOCALIZER)

    def test_text(self, sut: CopyrightNoticeT) -> None:
        """
        Tests :py:meth:`betty.copyright_notice.CopyrightNotice.text` implementations.
        """
        assert sut.text.localize(DEFAULT_LOCALIZER)

    def test_url(self, sut: CopyrightNoticeT) -> None:
        """
        Tests :py:meth:`betty.copyright_notice.CopyrightNotice.url` implementations.
        """
        url = sut.url
        if url is not None:
            assert url.localize(DEFAULT_LOCALIZER)



"""
JSON schemas for the date API.
"""

from __future__ import annotations

from typing import final

from betty.classtools import Singleton
from betty.json.linked_data import JsonLdObject
from betty.json.schema import Boolean, Null, Number, OneOf, String


@final
class DateSchema(Singleton, JsonLdObject):
    """
    A JSON Schema for :py:type:`betty.date.Date`.
    """

    def __init__(self):
        super().__init__(def_name="date", title="Date")
        self.add_property("fuzzy", Boolean(title="Fuzzy"))
        self.add_property("year", Number(title="Year"), False)
        self.add_property("month", Number(title="Month"), False)
        self.add_property("day", Number(title="Day"), False)
        self.add_property(
            "iso8601",
            String(
                pattern="^\\d\\d\\d\\d-\\d\\d-\\d\\d$", description="An ISO 8601 date."
            ),
            False,
        )


@final
class DateRangeSchema(Singleton, JsonLdObject):
    """
    A JSON Schema for :py:type:`betty.date.DateRange`.
    """

    def __init__(self):
        super().__init__(def_name="dateRange", title="Date range")
        date_schema = DateSchema()
        self._schema["additionalProperties"] = False
        self.add_property("start", OneOf(date_schema, Null(), title="Start date"))
        self.add_property("end", OneOf(date_schema, Null(), title="End date"))


@final
class ResolvableDateSchema(Singleton, OneOf):
    """
    A JSON Schema for :py:type:`betty.date.ResolvableDate`.
    """

    def __init__(self):
        super().__init__(
            DateSchema(),
            DateRangeSchema(),
            def_name="dateLike",
            title="Date or date range",
        )



"""Provide interactive family trees by integrating Betty with `Cytoscape.js <https://cytoscape.org/>`_."""

from __future__ import annotations

from pathlib import Path
from typing import TYPE_CHECKING, Self, final, override

from betty.extension import ExtensionDefinition
from betty.extension.trees.jobs import _GeneratePeopleJson
from betty.extension.webpack import Webpack
from betty.extension.webpack.build import EntryPointProvider
from betty.locale.localizable.gettext import _
from betty.project import Project
from betty.project.generate import Generator
from betty.service.factory import Manufacturable
from betty.service.requirement.project import require_project

if TYPE_CHECKING:
    from collections.abc import Sequence

    from betty.job.scheduler import Scheduler


@final
@ExtensionDefinition(
    "trees",
    label="Trees",
    description=_("Display interactive family trees using Cytoscape."),
    depends_on={Webpack},
    assets_directory=Path(__file__).parent / "assets",
)
class Trees(Generator, EntryPointProvider[Project], Manufacturable):
    """
    .. plugin:: extension:trees.
    """

    def __init__(self, *, project: Project):
        super().__init__()
        self._project = project

    @override
    @classmethod
    @require_project
    async def new(cls, project: Project, /) -> Self:
        return cls(project=project)

    @override
    async def generate(self, scheduler: Scheduler) -> None:
        await scheduler.add(_GeneratePeopleJson(project=self._project))

    @override
    @classmethod
    def webpack_entry_point_directory_path(cls) -> Path:
        return Path(__file__).parent / "webpack"

    @override
    def webpack_entry_point_cache_keys(self) -> Sequence[str]:
        return ()



"""Provide Betty's extension API."""

from __future__ import annotations

from typing import TYPE_CHECKING, final, override

from betty.definition.human_facing import HumanFacingDefinition
from betty.life_cycle.manage import ManagedLifeCycle
from betty.locale.localizable.gettext import _, ngettext
from betty.plugin import Plugin, PluginTypeDefinition, ResolvableId
from betty.plugin.dependent import DependentPluginDefinition
from betty.plugin.discovery.entry_point import EntryPointDiscovery
from betty.plugin.factory import PluginManufacturer
from betty.service.level import ServiceLevel

if TYPE_CHECKING:
    import builtins
    from collections.abc import Set
    from pathlib import Path

    from betty.locale.localizable import ResolvableLocalizable
    from betty.machine_name import ResolvableMachineName


class Extension[ServiceLevelT: ServiceLevel = ServiceLevel](
    ManagedLifeCycle, Plugin["ExtensionDefinition"]
):
    """
    Integrate custom services with a :py:class:`service level <betty.service.level.ServiceLevel>`.
    """


@final
@PluginTypeDefinition(
    "extension",
    label=_("Extension"),
    label_plural=_("Extensions"),
    label_countable=ngettext("{count} extension", "{count} extensions"),
    discovery=[EntryPointDiscovery("betty.extension")],
)
class ExtensionDefinition(HumanFacingDefinition, DependentPluginDefinition[Extension]):
    """
    .. plugin_type:: extension.

    Betty's functionality can be altered using *extensions*. An extension can do many things, such as loading new or
    expanding existing ancestry data, or generating additional content for your site.

    Some extensions are configurable. That means that other than enabling them, you can set the configuration options
    that determine how the extension should work. This can be done in your project's
    :py:class:`configuration <betty.project.data.ProjectConfiguration>`.
    """

    def __init__(
        self,
        plugin_id: ResolvableMachineName,
        *,
        label: ResolvableLocalizable,
        description: ResolvableLocalizable | None = None,
        comes_before: Set[ResolvableId] | None = None,
        comes_after: Set[ResolvableId] | None = None,
        depends_on: Set[ResolvableId] | None = None,
        assets_directory: Path | None = None,
        theme: bool = False,
    ):
        super().__init__(
            plugin_id,
            label=label,
            description=description,
            comes_before=comes_before,
            comes_after=comes_after,
            depends_on=depends_on,
        )
        self._assets_directory = assets_directory
        self._theme = theme

    @property
    def assets_directory(self) -> Path | None:
        """
        The path on disk where the extension's assets are located.
        """
        return self._assets_directory

    @property
    def theme(self) -> bool:
        """
        Whether this extension is a theme.
        """
        return self._theme


@final
class ExtensionManufacturer(PluginManufacturer[ExtensionDefinition, Extension]):
    """
    The extension manufacturer.
    """

    @override
    @classmethod
    def type(cls) -> builtins.type[ExtensionDefinition]:
        return ExtensionDefinition



"""
Test utilities for :py:mod:`betty.console`.
"""

from contextlib import redirect_stderr, redirect_stdout
from dataclasses import dataclass
from io import StringIO
from typing import final

from betty.app import App
from betty.console import SystemExitCode, main


@final
@dataclass
class Result:
    """
    A console run result.
    """

    exit_code: int
    stderr: str
    stdout: str


async def run(
    app: App, *args: str, expected_exit_code: SystemExitCode = SystemExitCode.OK
) -> Result:
    """
    Run a Betty console command.
    """
    stderr_f = StringIO()
    stdout_f = StringIO()
    with redirect_stderr(stderr_f), redirect_stdout(stdout_f):
        try:
            await main(app, args)
        except SystemExit as exception:
            if exception.code is None:
                exit_code = 0  # pragma: no cover
            elif isinstance(exception.code, int):
                exit_code = exception.code
            else:
                exit_code = 1  # pragma: no cover
        except BaseException as error:  # pragma: no cover
            raise AssertionError(f"The console did not raise {SystemExit}") from error
        else:  # pragma: no cover
            raise AssertionError(f"The console did not raise {SystemExit}")

    stderr_f.seek(0)
    stderr = stderr_f.read()
    stdout_f.seek(0)
    stdout = stdout_f.read()

    assert exit_code == expected_exit_code, f"""
The Betty command `{" ".join(args)}` unexpectedly exited with code {exit_code}, but {expected_exit_code} was expected.
Stdout:
{stdout}
Stderr:
{stderr}
"""
    return Result(exit_code, stderr, stdout)



from collections.abc import AsyncIterator, Sequence
from contextlib import asynccontextmanager
from datetime import datetime
from functools import partial
from typing import Any, Self, override

from betty.cache import Cache, CacheItem, CacheItemValueSetter
from betty.concurrent import AsynchronizedLock, Ledger
from betty.typing import threadsafe


class _StaticCacheItem[CacheItemValueT](CacheItem[CacheItemValueT]):
    __slots__ = "_value", "_modified"

    def __init__(
        self,
        value: CacheItemValueT,
        modified: int | float | None = None,
    ):
        self._value = value
        self._modified = datetime.now().timestamp() if modified is None else modified

    @override
    async def value(self) -> CacheItemValueT:
        return self._value

    @override
    @property
    def modified(self) -> int | float:
        return self._modified


class _CommonCacheBaseState[CacheT: Cache[Any]]:
    def __init__(
        self,
        cache_lock: AsynchronizedLock,
        cache_item_lock_ledger: Ledger,
    ):
        self.cache_lock = cache_lock
        self.cache_item_lock_ledger = cache_item_lock_ledger


@threadsafe
class _CommonCacheBase[CacheItemValueT](Cache[CacheItemValueT]):
    def __init__(
        self,
        *,
        scopes: Sequence[str] | None = None,
        state: _CommonCacheBaseState[Self] | None = None,
    ):
        self._scopes = scopes or ()
        if state is not None:
            self._cache_lock = state.cache_lock
            self._cache_item_lock_ledger = state.cache_item_lock_ledger
        else:
            self._cache_lock = AsynchronizedLock.new_threadsafe()
            self._cache_item_lock_ledger = Ledger(self._cache_lock)

    @override
    @asynccontextmanager
    async def hasset(
        self, cache_item_id: str, /
    ) -> AsyncIterator[CacheItemValueSetter[CacheItemValueT] | None]:
        if await self.has(cache_item_id):
            yield None
            return
        async with self._cache_item_lock_ledger.ledger(cache_item_id):
            if await self.has(cache_item_id):
                yield None
            yield partial(self.set, cache_item_id)
        return

    @override
    @asynccontextmanager
    async def getset(
        self, cache_item_id: str, /
    ) -> AsyncIterator[
        CacheItemValueSetter[CacheItemValueT] | CacheItem[CacheItemValueT]
    ]:
        if cache_item := await self.get(cache_item_id):
            yield cache_item
            return
        async with self._cache_item_lock_ledger.ledger(cache_item_id):
            if cache_item := await self.get(cache_item_id):
                yield cache_item
                return
            yield partial(self.set, cache_item_id)
        return



"""
Test utilities for :py:mod:`betty.locale.localizable`.
"""

from __future__ import annotations

from typing import TYPE_CHECKING

from betty.locale import DEFAULT_LOCALE
from betty.locale.localizable.plain import Plain
from betty.locale.localizable.static import CountableStaticTranslations

if TYPE_CHECKING:
    from betty.locale.localizable import CountableLocalizable, Localizable

DUMMY_LOCALIZABLE: Localizable = Plain("DUMMY_LOCALIZABLE")
"""
A dummy localizable.
"""

DUMMY_COUNTABLE_LOCALIZABLE: CountableLocalizable = CountableStaticTranslations(
    {
        DEFAULT_LOCALE: {
            "one": "{count} DUMMY_COUNTABLE_LOCALIZABLE",
            "other": "{count} DUMMY_COUNTABLE_LOCALIZABLES",
        }
    }
)
"""
A dummy countable localizable.
"""



"""
Generic plugin API errors.
"""

from __future__ import annotations

from typing import TYPE_CHECKING, Any, final

from betty.exception import HumanFacingException
from betty.locale.localizable.gettext import _
from betty.locale.localizable.markup import Paragraph, do_you_mean
from betty.plugin import (
    PluginDefinition,
    PluginTypeDefinition,
    ResolvableId,
    resolve_id,
)

if TYPE_CHECKING:
    from collections.abc import Sequence

    from betty.machine_name import ResolvableMachineName


class PluginError(Exception):
    """
    Any error originating from the Plugin API.
    """


class PluginUnavailable(PluginError, HumanFacingException):
    """
    Raised when a plugin is unavailable for use.
    """


@final
class PluginNotFound(PluginUnavailable):
    """
    Raised when a plugin cannot be found.
    """

    def __init__[PluginDefinitionT: PluginDefinition](
        self,
        plugin_type: PluginTypeDefinition[Any, PluginDefinitionT],
        plugin_not_found: ResolvableMachineName,
        available_plugins: Sequence[ResolvableId[PluginDefinitionT]],
        /,
    ):
        super().__init__(
            Paragraph(
                _('Could not find a(n) {plugin_type} plugin "{plugin_id}".').format(
                    plugin_type=plugin_type.label, plugin_id=plugin_not_found
                ),
                do_you_mean(
                    *[
                        f'"{resolve_id(available_plugin)}"'
                        for available_plugin in available_plugins
                    ]
                ),
            )
        )



"""
Render plain text to HTML.
"""

from typing import override

from betty.html import plain_text_to_html
from betty.locale.localizable.gettext import _
from betty.media_type import MediaType
from betty.media_type.media_types import PLAIN_TEXT
from betty.render import Renderer, RendererDefinition


@RendererDefinition("plain-text", label=_("Plain text"))
class PlainText(Renderer):
    """
    .. plugin:: renderer:plain-text.
    """

    @override
    @property
    def media_type(self) -> MediaType:
        return PLAIN_TEXT

    @override
    async def render(self, content: str, /) -> str:
        return plain_text_to_html(content)



"""
Provide the place entity.
"""

from __future__ import annotations

from contextlib import suppress
from typing import TYPE_CHECKING, final, override

from betty.ancestry.has_file_references import HasFileReferences
from betty.ancestry.has_links import HasLinks
from betty.ancestry.has_notes import HasNotes
from betty.ancestry.name import Name
from betty.json.linked_data import JsonLdObject, dump_context
from betty.json.schema import Array, Number, Object
from betty.locale.localizable.gettext import _, ngettext
from betty.model import EntityDefinition
from betty.model.association import BidirectionalToManySingleType, ToManyAssociates
from betty.place_type.place_types import Unknown as UnknownPlaceType
from betty.privacy import HasPrivacy

if TYPE_CHECKING:
    from collections.abc import MutableSequence

    from geopy import Point

    from betty.ancestry.enclosure import Enclosure
    from betty.ancestry.event import Event
    from betty.ancestry.link import Link
    from betty.ancestry.note import Note
    from betty.locale.localizable import Localizable
    from betty.place_type import PlaceType
    from betty.portable import PortableMapping
    from betty.privacy import Privacy
    from betty.project import Project


@final
@EntityDefinition(
    "place",
    label=_("Place"),
    label_plural=_("Places"),
    label_countable=ngettext("{count} place", "{count} places"),
)
class Place(HasLinks, HasFileReferences, HasNotes, HasPrivacy):
    """
    .. plugin:: entity:place.
    """

    events = BidirectionalToManySingleType["Place", "Event"](
        "betty.ancestry.event:Event",
        "place",
        label=_("Events"),
        description=_("The events that happened in this place"),
    )
    """
    The events that happened here.
    """

    enclosers = BidirectionalToManySingleType["Place", "Enclosure"](
        "betty.ancestry.enclosure:Enclosure",
        "enclosee",
        label=_("Enclosers"),
        description=_("The places this place is enclosed or contained by"),
        linked_data_embedded=True,
    )
    """
    Other places containing this one.
    """

    enclosees = BidirectionalToManySingleType["Place", "Enclosure"](
        "betty.ancestry.enclosure:Enclosure",
        "encloser",
        label=_("Enclosees"),
        description=_("The places this place encloses or contains"),
        linked_data_embedded=True,
    )
    """
    Other places contained by this one.
    """

    def __init__(
        self,
        *,
        id: str | None = None,  # noqa: A002
        names: MutableSequence[Name] | None = None,
        events: ToManyAssociates[Event] | None = None,
        enclosers: ToManyAssociates[Enclosure] | None = None,
        enclosees: ToManyAssociates[Enclosure] | None = None,
        notes: ToManyAssociates[Note] | None = None,
        coordinates: Point | None = None,
        links: MutableSequence[Link] | None = None,
        privacy: Privacy | None = None,
        place_type: PlaceType | None = None,
    ):
        super().__init__(id, notes=notes, links=links, privacy=privacy)
        self._names = [] if names is None else names
        self._coordinates = coordinates
        if events is not None:
            self.events = events
        if enclosers is not None:
            self.enclosers = enclosers
        if enclosees is not None:
            self.enclosees = enclosees
        self._place_type = place_type or UnknownPlaceType()

    @property
    def place_type(self) -> PlaceType:
        """
        The type of this place.
        """
        return self._place_type

    @place_type.setter
    def place_type(self, place_type: PlaceType) -> None:
        self._place_type = place_type

    @property
    def names(self) -> MutableSequence[Name]:
        """
        The place's names.

        The first name is considered the :py:attr:`place label <betty.ancestry.place.Place.label>`.
        """
        return self._names

    @property
    def coordinates(self) -> Point | None:
        """
        The place's coordinates.
        """
        return self._coordinates

    @coordinates.setter
    def coordinates(self, coordinates: Point):
        self._coordinates = coordinates

    @override
    @property
    def label(self) -> Localizable:
        with suppress(IndexError):
            return self.names[0].name
        return super().label

    @override
    async def dump_linked_data(self, project: Project, /) -> PortableMapping:
        portable = await super().dump_linked_data(project)
        dump_context(
            portable,
            names="https://schema.org/name",
            events="https://schema.org/event",
            enclosers="https://schema.org/containedInPlace",
            enclosees="https://schema.org/containsPlace",
        )
        portable["@type"] = "https://schema.org/Place"
        portable["names"] = [
            await name.dump_linked_data(project) for name in self.names
        ]
        if self.coordinates is not None:
            portable_coordinates: PortableMapping = {
                "@type": "https://schema.org/GeoCoordinates",
                "latitude": self.coordinates.latitude,
                "longitude": self.coordinates.longitude,
            }
            dump_context(portable, coordinates="https://schema.org/geo")
            dump_context(portable_coordinates, latitude="https://schema.org/latitude")
            dump_context(portable_coordinates, longitude="https://schema.org/longitude")
            portable["coordinates"] = portable_coordinates
        return portable

    @override
    @classmethod
    async def linked_data_schema(cls, project: Project, /) -> JsonLdObject:
        schema = await super().linked_data_schema(project)
        schema.add_property(
            "names", Array(await Name.linked_data_schema(project), title="Names")
        )
        coordinate_schema = Number(title="Coordinate")
        coordinates_schema = Object(title="Coordinates")
        coordinates_schema.add_property("latitude", coordinate_schema, False)
        coordinates_schema.add_property("longitude", coordinate_schema, False)
        schema.add_property("coordinates", coordinates_schema, False)
        return schema



"""
Provide utilities for running jobs concurrently.
"""

from __future__ import annotations

from abc import ABC, abstractmethod
from datetime import datetime
from typing import TYPE_CHECKING, Any, final
from uuid import uuid4

from betty.cache.memory import MemoryCache
from betty.progress.no_op import NoOpProgress

if TYPE_CHECKING:
    from betty.cache import Cache
    from betty.job.scheduler import Scheduler
    from betty.progress import Progress


@final
class Context:
    """
    A job context.
    """

    def __init__(self, *, progress: Progress | None = None):
        self._id = str(uuid4())
        self._cache = MemoryCache()
        self._start = datetime.now()
        self._progress = progress or NoOpProgress()

    @property
    def id(self) -> str:
        """
        The unique job context ID.
        """
        return self._id

    @property
    def cache(self) -> Cache[Any]:
        """
        Provide a cache for this job context.

        The cache is volatile and will be discarded once the job context is completed.
        """
        return self._cache

    @property
    def start(self) -> datetime:
        """
        When the job started.
        """
        return self._start

    @property
    def progress(self) -> Progress:
        """
        The job progress.
        """
        return self._progress


class Job(ABC):
    """
    A job.
    """

    def __init__(
        self,
        job_id: str,
        *,
        dependencies: set[str] | None = None,
        dependents: set[str] | None = None,
        priority: bool = False,
    ):
        self._called = False
        self._id = job_id
        self._dependencies = set() if dependencies is None else dependencies
        self._dependents = set() if dependents is None else dependents
        self._priority = priority

    @property
    def id(self) -> str:
        """
        The unique job ID.
        """
        return self._id

    @property
    def dependencies(self) -> set[str]:
        """
        The IDs of any other jobs this job depends on.
        """
        return self._dependencies

    @property
    def dependents(self) -> set[str]:
        """
        The IDs of any other jobs that depend on this job.
        """
        return self._dependents

    @property
    def priority(self) -> bool:
        """
        Whether the job has priority over others.
        """
        return self._priority

    @abstractmethod
    async def do(self, scheduler: Scheduler, /) -> None:
        """
        Do the job.
        """



"""
Common serializers.
"""

from __future__ import annotations

import json
from collections.abc import Mapping, Sequence
from typing import TYPE_CHECKING, cast, final, override

import yaml

from betty.locale.localizable.gettext import _
from betty.media_type.media_types import JSON, YAML
from betty.portable import PortableData
from betty.serde import SerializationError, Serializer, SerializerDefinition

if TYPE_CHECKING:
    from betty.media_type import MediaType


@final
@SerializerDefinition("json", label="JSON")
class Json(Serializer):
    """
    .. plugin:: serializer:json.
    """

    @override
    @classmethod
    def media_type(cls) -> MediaType:
        return JSON

    @override
    def load(self, serialized: str, /) -> PortableData:
        try:
            return cast(PortableData, json.loads(serialized))
        except json.JSONDecodeError as e:
            raise SerializationError(
                _("Invalid JSON: {error}.").format(error=str(e))
            ) from None

    @override
    def dump(self, portable: PortableData, /) -> str:
        return json.dumps(portable, indent=2, sort_keys=True)


@final
@SerializerDefinition("yaml", label="YAML")
class Yaml(Serializer):
    """
    .. plugin:: serializer:yaml.
    """

    @override
    @classmethod
    def media_type(cls) -> MediaType:
        return YAML

    @override
    def load(self, serialized: str, /) -> PortableData:
        try:
            return cast(PortableData, yaml.safe_load(serialized))
        except yaml.YAMLError as e:
            raise SerializationError(
                _("Invalid YAML: {error}.").format(error=str(e))
            ) from None

    @override
    def dump(self, portable: PortableData, /) -> str:
        return yaml.safe_dump(self._safe_str(portable))

    def _safe_str(self, value: PortableData) -> PortableData:
        # Work around a bug where ``str`` subclasses cannot be serialized to YAML.
        if isinstance(value, str):
            return str(value)
        if isinstance(value, Mapping):
            return dict(
                zip(
                    map(self._safe_str, value.keys()),  # ty:ignore[invalid-argument-type]
                    map(self._safe_str, value.values()),  # ty:ignore[invalid-argument-type]
                    strict=False,
                )
            )
        if isinstance(value, Sequence):
            return list(map(self._safe_str, value))  # ty:ignore[invalid-argument-type]
        return value



"""
Jobs.
"""

from __future__ import annotations

from typing import TYPE_CHECKING, override

from betty.job import Job

if TYPE_CHECKING:
    from pathlib import Path

    from betty.gramps.loader import GrampsLoader
    from betty.job.scheduler import Scheduler


class LoadAncestry(Job):
    """
    Load Gramps data into an ancestry.
    """

    def __init__(self, *, loader: GrampsLoader, source: Path | str):
        super().__init__(self.id_for(source))
        self._loader = loader
        self._source = source

    @classmethod
    def id_for(cls, source: Path | str, /) -> str:
        """
        Get the job ID.
        """
        return f"gramps:load-ancestry:{id(type(source))}:{str(source)}"

    @override
    async def do(self, scheduler: Scheduler, /) -> None:
        if isinstance(self._source, str):
            await self._loader.load_name(self._source)
        else:
            await self._loader.load_file(self._source)



"""
Event type configuration.
"""

from __future__ import annotations

from typing import final, override

from betty.data.aggregate.record.object import ObjectDefinition
from betty.event_type import EventType, EventTypeDefinition
from betty.locale import DEFAULT_LOCALE
from betty.locale.localizable.gettext import _
from betty.locale.localizable.static import CountableStaticTranslations
from betty.plugin.data import CountableHumanFacingPluginDefinitionConfiguration
from betty.plugin.data.ordered import OrderedPluginDefinitionConfiguration
from betty.sample import Sample


@final
@ObjectDefinition(
    label=_("Event type configuration"),
    samples=[
        lambda: Sample(
            EventTypeDefinitionConfiguration(
                id="moon-landing",
                label="Moon landing",
                label_plural="Moon landings",
                label_countable=CountableStaticTranslations(
                    {
                        DEFAULT_LOCALE: {
                            "one": "{count} moon landing",
                            "other": "{count} moon landings",
                        }
                    }
                ),
            ),
            label="Default",
        )
    ],
)
class EventTypeDefinitionConfiguration(
    CountableHumanFacingPluginDefinitionConfiguration[EventTypeDefinition],
    OrderedPluginDefinitionConfiguration[EventTypeDefinition],
):
    """
    Configure a :py:class:`betty.event_type.EventTypeDefinition`.

    .. data:: betty.project.data:EventTypeDefinitionConfiguration
    """

    @override
    def new_plugin(self) -> EventTypeDefinition:
        @EventTypeDefinition(
            self.id,
            label=self.label,
            label_plural=self.label_plural,
            label_countable=self.label_countable,
            description=self.description,
            comes_before=set(self.comes_before),
            comes_after=set(self.comes_after),
        )
        class _ProjectConfigurationEventType(EventType):
            pass

        return _ProjectConfigurationEventType.plugin()



"""
Test utilities for :py:mod:`betty.job.scheduler`.
"""

from __future__ import annotations

from asyncio import create_task, sleep
from collections.abc import Callable, Iterator, MutableSequence
from typing import TYPE_CHECKING, cast, final, override

import pytest

from betty.job import Context, Job
from betty.job.executor.asyncio import AsyncExecutor
from betty.job.scheduler import (
    Cancelled,
    Closed,
    Completed,
    CyclicDependencyError,
    DuplicateJobError,
    Released,
    ScheduledJobBatch,
    Scheduler,
    UnknownJobError,
)

if TYPE_CHECKING:
    from collections.abc import Iterable, Sequence


class StaticScheduler(Scheduler):
    """
    A scheduler that issues static job batches.
    """

    def __init__(self, context: Context, batches: Sequence[ScheduledJobBatch | Closed]):
        super().__init__(context)
        self._batches: Iterable[ScheduledJobBatch | Closed] | Closed = iter(batches)

    @override
    async def get(self) -> ScheduledJobBatch:
        if isinstance(self._batches, Closed):
            raise self._batches
        try:
            batch = next(cast(Iterator[ScheduledJobBatch | Closed], self._batches))
        except StopIteration:
            self._batches = Completed()
            raise self._batches from None
        else:
            if isinstance(batch, Closed):
                self._batches = batch
                raise batch
            return batch

    @override
    async def add(self, *jobs: Job) -> None:
        raise NotImplementedError

    @override
    async def release(self) -> None:
        raise NotImplementedError

    @override
    async def cancel(self, reason: BaseException | None = None, /) -> None:
        raise NotImplementedError

    @override
    async def complete(self) -> None:
        while not isinstance(self._batches, Closed):
            await sleep(0)
        if not isinstance(self._batches, Completed):
            raise self._batches

    @override
    @property
    def context(self) -> Context:
        return self._context


class _Job(Job):
    def __init__(
        self,
        job_id: str,
        carrier: MutableSequence[_Job] | None = None,
        *,
        additional_jobs: Sequence[Callable[[MutableSequence[_Job]], _Job]]
        | None = None,
        dependencies: set[str] | None = None,
        dependents: set[str] | None = None,
        priority: bool = False,
    ):
        super().__init__(
            job_id, dependencies=dependencies, dependents=dependents, priority=priority
        )
        self.carrier = [] if carrier is None else carrier
        self._additional_jobs = additional_jobs

    @override
    async def do(self, scheduler: Scheduler, /) -> None:
        if self._additional_jobs is not None:
            for additional_job in self._additional_jobs:
                await scheduler.add(additional_job(self.carrier))
        if self.carrier is not None:
            self.carrier.append(self)


@final
class Sleep(Job):
    """
    A job that sleeps for a long, long time.
    """

    @override
    async def do(self, scheduler: Scheduler, /) -> None:
        await sleep(999999999)


class Raise(Job):
    """
    A job that raises an exception.
    """

    def __init__(
        self,
        job_id: str,
        *,
        reason: BaseException,
        dependencies: set[str] | None = None,
        dependents: set[str] | None = None,
        priority: bool = False,
    ):
        super().__init__(
            job_id, dependencies=dependencies, dependents=dependents, priority=priority
        )
        self._reason = reason

    @override
    async def do(self, scheduler: Scheduler, /) -> None:
        raise self._reason


class SchedulerTestBase:
    """
    A base class for testing :py:class:`betty.job.scheduler.Scheduler` implementations.
    """

    @pytest.fixture
    def sut(self) -> Scheduler:
        """
        Provide the systems under test.
        """
        raise NotImplementedError

    async def test___aexit___with_exception(self, sut: Scheduler) -> None:
        """
        Tests :py:meth:`betty.job.scheduler.Scheduler.__aexit__` implementations.
        """
        exception = RuntimeError()
        with pytest.raises(type(exception)) as exc_info:
            async with sut:
                raise exception
        assert exc_info.value is exception

    async def test_add(self, sut: Scheduler) -> None:
        """
        Tests :py:meth:`betty.job.scheduler.Scheduler.add` implementations.
        """
        await sut.add(_Job("job"))

    async def test_add__with_duplicate_job(self, sut: Scheduler) -> None:
        """
        Tests :py:meth:`betty.job.scheduler.Scheduler.add` implementations.
        """
        await sut.add(_Job("job"))
        with pytest.raises(DuplicateJobError):
            await sut.add(_Job("job"))

    async def test_add__with_dependent_post_release(self, sut: Scheduler) -> None:
        """
        Tests :py:meth:`betty.job.scheduler.Scheduler.add` implementations.
        """
        await sut.release()
        with pytest.raises(Released):
            await sut.add(
                _Job("dependent"), _Job("dependency", dependents={"dependent"})
            )

    async def test_release(self, sut: Scheduler) -> None:
        """
        Tests :py:meth:`betty.job.scheduler.Scheduler.release` implementations.
        """
        await sut.release()
        await sut.complete()

    @pytest.mark.parametrize(
        ("expected", "pre_release_jobs", "post_release_jobs"),
        [
            # No jobs.
            ([], [], []),
            # Isolated jobs.
            (
                ["isolated"],
                [],
                [
                    lambda carrier: _Job("isolated", carrier),
                ],
            ),
            (
                {f"isolated:{index}" for index in range(999)},
                [],
                [
                    lambda carrier, index=index: _Job(f"isolated:{index}", carrier)
                    for index in range(999)
                ],
            ),
            # A simple dependency, provided in different orders.
            (
                ["dependency", "dependent"],
                [],
                [
                    lambda carrier: _Job("dependency", carrier),
                    lambda carrier: _Job(
                        "dependent", carrier, dependencies={"dependency"}
                    ),
                ],
            ),
            (
                ["dependency", "dependent"],
                [],
                [
                    lambda carrier: _Job(
                        "dependent", carrier, dependencies={"dependency"}
                    ),
                    lambda carrier: _Job("dependency", carrier),
                ],
            ),
            # A simple dependent.
            (
                ["dependency", "dependent"],
                [
                    lambda carrier: _Job(
                        "dependency", carrier, dependents={"dependent"}
                    ),
                ],
                [
                    lambda carrier: _Job("dependent", carrier),
                ],
            ),
            # One job adding another, isolated job
            (
                ["one", "other"],
                [],
                [
                    lambda carrier: _Job(
                        "one",
                        carrier,
                        additional_jobs=[lambda carrier: _Job("other", carrier)],
                    )
                ],
            ),
            # Diamond-shaped dependencies.
            (
                {"one", "two-one", "two-two", "three"},
                [],
                [
                    lambda carrier: _Job("one", carrier),
                    lambda carrier: _Job("two-one", carrier, dependencies={"one"}),
                    lambda carrier: _Job("two-two", carrier, dependencies={"one"}),
                    lambda carrier: _Job(
                        "three", carrier, dependencies={"two-one", "two-two"}
                    ),
                ],
            ),
            # One job adding another job as a dependent by being its dependency.
            (
                ["dependency", "dependent"],
                [],
                [
                    lambda carrier: _Job(
                        "dependency",
                        carrier,
                        additional_jobs=[
                            lambda carrier: _Job(
                                "dependent", carrier, dependencies={"dependency"}
                            )
                        ],
                    )
                ],
            ),
            # Multiple jobs that must be run in a different order due to durations.
            (
                ["priority", "no-priority"],
                [],
                [
                    lambda carrier: _Job("no-priority", carrier, priority=False),
                    lambda carrier: _Job("priority", carrier, priority=True),
                ],
            ),
        ],
    )
    async def test_get__returns_job(
        self,
        expected: set[str] | Sequence[str],
        pre_release_jobs: Sequence[Callable[[MutableSequence[_Job]], _Job]],
        post_release_jobs: Sequence[Callable[[MutableSequence[_Job]], _Job]],
        sut: Scheduler,
    ) -> None:
        """
        Tests :py:meth:`betty.job.scheduler.Scheduler.get` implementations.
        """
        carrier = []
        async with AsyncExecutor(sut):
            await sut.add(*(job(carrier) for job in pre_release_jobs))
            await sut.release()
            await sut.add(*(job(carrier) for job in post_release_jobs))
            await sut.complete()

        actual_ids = (job.id for job in carrier)
        actual = set(actual_ids) if isinstance(expected, set) else list(actual_ids)
        assert actual == expected

    async def test_get__raises_when_unknown_dependency(
        self,
        sut: Scheduler,
    ) -> None:
        """
        Tests :py:meth:`betty.job.scheduler.Scheduler.get` implementations.
        """
        async with AsyncExecutor(sut):
            await sut.add(_Job("dependent", dependencies={"dependency"}))
            with pytest.raises(UnknownJobError):
                async with sut:
                    pass

    async def test_get__raises_when_cyclic(self, sut: Scheduler) -> None:
        """
        Tests :py:meth:`betty.job.scheduler.Scheduler.get` implementations.
        """
        await sut.add(_Job("one", dependencies={"other"}))
        await sut.add(_Job("other", dependencies={"one"}))
        with pytest.raises(CyclicDependencyError):
            async with sut:
                await sut.get()

    async def test_get__raises_when_cancelled(self, sut: Scheduler) -> None:
        """
        Tests :py:meth:`betty.job.scheduler.Scheduler.get` implementations.
        """
        await sut.cancel()
        with pytest.raises(Cancelled):
            await sut.get()

    async def test_get__raises_when_completed(self, sut: Scheduler) -> None:
        """
        Tests :py:meth:`betty.job.scheduler.Scheduler.get` implementations.
        """
        await sut.complete()
        with pytest.raises(Completed):
            await sut.get()

    async def test_get___job_exception_should_cancel(self, sut: Scheduler) -> None:
        """
        Tests :py:meth:`betty.job.scheduler.Scheduler.get` implementations.
        """
        reason = RuntimeError()
        await sut.add(Raise("", reason=reason))
        with pytest.raises(Cancelled) as exc_info:  # noqa: PT012
            async with sut:
                batch = await sut.get()
                await batch()
        assert exc_info.value.__cause__ is reason

    async def test_get___done_when_cancelled(self, sut: Scheduler) -> None:
        """
        Tests :py:meth:`betty.job.scheduler.Scheduler.get` implementations.
        """
        carrier = []
        job = _Job("one", carrier)
        await sut.add(job)
        with pytest.raises(Cancelled):  # noqa: PT012
            async with sut:
                batch = await sut.get()
                await sut.cancel()
                await batch()

        assert carrier == [job]

    async def test___aiter__(self, sut: Scheduler) -> None:
        """
        Tests :py:meth:`betty.job.scheduler.Scheduler.__aiter__` implementations.
        """
        completion = create_task(sut.complete())
        assert [batch async for batch in sut] == []
        await completion

    async def test_complete__when_completed(self, sut: Scheduler) -> None:
        """
        Tests :py:meth:`betty.job.scheduler.Scheduler.complete` implementations.
        """
        await sut.complete()
        await sut.complete()

    async def test_complete__when_cancelled(self, sut: Scheduler) -> None:
        """
        Tests :py:meth:`betty.job.scheduler.Scheduler.complete` implementations.
        """
        await sut.cancel()
        with pytest.raises(Cancelled):
            await sut.complete()

    async def test_cancel(self, sut: Scheduler) -> None:
        """
        Tests :py:meth:`betty.job.scheduler.Scheduler.cancel` implementations.
        """
        await sut.add(Sleep(""))
        await sut.cancel()
        with pytest.raises(Cancelled):
            await sut.get()

    async def test_cancel__with_exception(self, sut: Scheduler) -> None:
        """
        Tests :py:meth:`betty.job.scheduler.Scheduler.cancel` implementations.
        """
        reason = RuntimeError()

        await sut.add(Sleep(""))
        await sut.cancel(reason)
        with pytest.raises(Cancelled) as exc_info:
            await sut.get()
        assert exc_info.value.__cause__ is reason

    async def test_cancel__when_cancelled(self, sut: Scheduler) -> None:
        """
        Tests :py:meth:`betty.job.scheduler.Scheduler.cancel` implementations.
        """
        await sut.cancel()
        await sut.cancel()

    async def test_cancel__when_completed(self, sut: Scheduler) -> None:
        """
        Tests :py:meth:`betty.job.scheduler.Scheduler.cancel` implementations.
        """
        await sut.complete()
        await sut.cancel()

    async def test_context(self, sut: Scheduler) -> None:
        """
        Tests :py:meth:`betty.job.scheduler.Scheduler.context` implementations.
        """
        assert sut.context is sut.context



"""
Job execution using thread pools.
"""

from __future__ import annotations

import asyncio
from asyncio import Task, gather, get_running_loop
from concurrent import futures
from contextlib import suppress
from typing import TYPE_CHECKING, final, override

from betty.job.executor import Executor
from betty.job.executor.asyncio import AsyncExecutor
from betty.job.scheduler import Cancelled, Scheduler

if TYPE_CHECKING:
    from collections.abc import MutableSequence
    from concurrent.futures import Future


@final
class ThreadPoolExecutor(Executor):
    """
    A job executor using a thread pool.
    """

    def __init__(
        self,
        scheduler: Scheduler,
        *,
        async_concurrency: int = 1,
        threading_concurrency: int = 1,
    ):
        assert async_concurrency > 0
        assert threading_concurrency > 0
        self._scheduler = scheduler
        self._async_concurrency = async_concurrency
        self._threading_concurrency = threading_concurrency
        self._working = False
        self._futures: set[Future[None]] = set()
        self._tasks: set[Task[None]] = set()
        self._thread_pool = futures.ThreadPoolExecutor(
            max_workers=threading_concurrency
        )
        self._event_loop = get_running_loop()
        self._async_executors: MutableSequence[Executor] = []

    @override
    async def start(self) -> None:
        if self._working:
            return
        self._working = True
        for _ in range(self._threading_concurrency):
            self._futures.add(self._thread_pool.submit(self._run_job))

    def _run_job(self) -> None:
        self._tasks.add(self._event_loop.create_task(self.__run_job()))

    async def __run_job(self) -> None:
        async_executor = AsyncExecutor(
            self._scheduler, concurrency=self._async_concurrency
        )
        self._async_executors.append(async_executor)
        try:
            async with async_executor:
                pass
        except Cancelled:
            await self.cancel()

    @override
    async def cancel(self) -> None:
        self._working = False
        await gather(
            *[async_executor.cancel() for async_executor in self._async_executors]
        )
        for task in self._tasks:
            task.cancel()
        for future in self._futures:
            future.cancel()
        self._thread_pool.shutdown(False, cancel_futures=True)

    @override
    async def complete(self) -> None:
        if not self._working:
            return
        await gather(
            *[async_executor.complete() for async_executor in self._async_executors]
        )
        for task in asyncio.as_completed(self._tasks):
            with suppress(asyncio.CancelledError):
                await task
        for future in futures.as_completed(self._futures):
            with suppress(futures.CancelledError):
                future.result()
        self._thread_pool.shutdown()
        self._working = False



"""
Key-value mapping data types.
"""

from __future__ import annotations

from collections.abc import Callable, MutableMapping
from typing import TYPE_CHECKING, Any, final

from betty.data import DataDefinition
from betty.data.aggregate.collection import CollectionDefinition
from betty.data.indicator.selector import Key
from betty.portable import CallbackPorter, PortableData, Porter

if TYPE_CHECKING:
    from ty_extensions import Intersection

    from betty.data import Data
    from betty.locale.localizable import ResolvableLocalizable


@final
class MappingDefinition[MutableMappingT: MutableMapping[Any, Any]](
    CollectionDefinition[MutableMappingT, Key]
):
    """
    A key-value mapping data definition.
    """

    def __init__[KeyT, ValueT](
        self,
        /,
        cls: type[Intersection[MutableMappingT, MutableMapping[KeyT, ValueT]]]
        | None = None,
        *,
        key: DataDefinition[KeyT, str] | type[Intersection[KeyT, Data]],
        value: DataDefinition[ValueT] | type[Intersection[ValueT, Data]],
        label: ResolvableLocalizable,
        description: ResolvableLocalizable | None = None,
        factory: Callable[[], MutableMappingT] | None = None,
        porter: Porter[MutableMappingT] | None = None,
    ):
        super().__init__(
            cls=cls,
            item=key,
            label=label,
            description=description,
            porter=CallbackPorter(self._load, self._dump) if porter is None else porter,
        )
        self._value = value if isinstance(value, DataDefinition) else value.data()
        self._factory = factory

    def _load(self, portable: PortableData, /) -> MutableMappingT:
        from betty.assertion import assert_mapping

        loaded = self.new()
        loaded.update(
            assert_mapping(self._value.porter.load, self._item.porter.load)(portable)
        )
        return loaded

    def _dump(self, data: MutableMappingT) -> PortableData:
        return {
            self._item.porter.dump(key): self._value.porter.dump(item)
            for key, item in data.items()
        }  # ty:ignore[invalid-return-type]



"""
Jobs.
"""

from __future__ import annotations

from typing import TYPE_CHECKING, override

from betty.deriver import Deriver as DeriverApi
from betty.job import Job

if TYPE_CHECKING:
    from betty.job.scheduler import Scheduler
    from betty.project import Project


class DeriveAncestry(Job):
    """
    Derive information for an ancestry.
    """

    def __init__(self, *, project: Project):
        super().__init__(self.id_for())
        self._project = project

    @classmethod
    def id_for(cls) -> str:
        """
        Get the job ID.
        """
        return "deriver:derive"

    @override
    async def do(self, scheduler: Scheduler, /) -> None:
        deriver = DeriverApi(self._project)
        await deriver.derive()



"""
Keyed collection types and implementations.
"""

from abc import abstractmethod
from collections.abc import Collection, Iterable

from betty.collection import MutableCollection


class KeyedCollection[KeyT, ResolvableKeyT, ValueT](Collection[ValueT]):
    """
    A collection of values that are accessible by their keys.

    This is different from a mapping in that iteration takes place over values, not keys, and that callers cannot
    necessarily associate keys with values themselves (this may be done automatically by implementations).
    """

    @abstractmethod
    def keys(self) -> Iterable[KeyT]:
        """
        Get an iterable over the collection's keys.
        """

    @abstractmethod
    def __getitem__(self, key: KeyT | ResolvableKeyT) -> ValueT:
        pass


class MutableKeyedCollection[KeyT, ResolvableKeyT, ValueT, ResolvableValueT](
    KeyedCollection[KeyT, ResolvableKeyT, ValueT],
    MutableCollection[ValueT],
):
    """
    A mutable ordered collection of values that are accessible by their keys.
    """

    @abstractmethod
    def add(self, *values: ValueT | ResolvableValueT) -> None:
        """
        Add one or more values to the collection.
        """

    @abstractmethod
    def remove(self, *keys: KeyT | ResolvableKeyT) -> None:
        """
        Remove one or more keys from the collection.
        """

    @abstractmethod
    def __delitem__(self, key: KeyT | ResolvableKeyT) -> None:
        pass



"""
Sequence properties.
"""

from __future__ import annotations

from collections.abc import Callable, Iterable, MutableSequence
from typing import TYPE_CHECKING, Any, override

from betty.functools import passthrough
from betty.property import Property

if TYPE_CHECKING:
    from collections.abc import Callable

    from ty_extensions import Intersection

    from betty.data import Data, DataDefinition
    from betty.data.aggregate.collection.sequence import SequenceDefinition
    from betty.locale.localizable import ResolvableLocalizable


class SequenceProperty[MutableSequenceT: MutableSequence[Any], ValueSetT](
    Property[MutableSequenceT, ValueSetT]
):
    """
    A property that contains a :py:class:`collections.abc.MutableSequence`.
    """

    _data: SequenceDefinition[MutableSequenceT]

    def __init__(
        self,
        data: Intersection[DataDefinition[MutableSequenceT], SequenceDefinition]
        | Data[Intersection[DataDefinition[MutableSequenceT], SequenceDefinition]],
        *,
        label: ResolvableLocalizable | None = None,
        description: ResolvableLocalizable | None = None,
        omit_load: bool | None = None,
        omit_dump: Callable[[MutableSequenceT], bool] | None = None,
        resolver: Callable[
            [ValueSetT | Iterable[ValueSetT]], Iterable[ValueSetT]
        ] = passthrough,
        default: Callable[[], ValueSetT | Iterable[ValueSetT]] = list,
    ):
        super().__init__(
            data,
            label=label,
            description=description,
            omit_load=omit_load,
            omit_dump=omit_dump,
            default=self._new_default,
        )
        self._values_resolver = resolver
        self._default_values = default

    def _new_default(self) -> MutableSequenceT:
        new = self._data.new()
        new.extend(self._values_resolver(self._default_values()))
        return new

    @override
    def set(
        self, instance: Any, value: ValueSetT | MutableSequenceT
    ) -> MutableSequenceT:
        data = self.get(instance)
        data.clear()
        data.extend(self._values_resolver(value))
        return data



"""
Provide the project API.

Projects are how people use Betty. A project is a workspace, starting out with the user's configuration,
and combining it with the resulting ancestry, allowing the user to perform tasks, such as generating a
site from the entire project.
"""

from __future__ import annotations

from contextlib import AsyncExitStack, asynccontextmanager
from pathlib import Path
from typing import TYPE_CHECKING, Self, final, overload, override

from aiofiles.tempfile import TemporaryDirectory

import betty
import betty.dirs
from betty.ancestry import Ancestry
from betty.asset import AssetRepository, ProxyAssetRepository, StaticAssetRepository
from betty.document import Document, DocumentProvider
from betty.extension import Extension, ExtensionDefinition
from betty.hashid import hashid
from betty.importlib import fully_qualified_name
from betty.life_cycle.manage import ManagedLifeCycle
from betty.locale import DEFAULT_LOCALE
from betty.locale.localizable.gettext import _
from betty.locale.localize import Localizer, LocalizerRepository
from betty.locale.translation import (
    AssetTranslationRepository,
    ProxyTranslationRepository,
    TranslationRepository,
)
from betty.machine_name import MachineName
from betty.plugin import ResolvableId, resolve_id
from betty.plugin.dependent import sort_dependent_plugin_graph
from betty.privacy.privatizer import Privatizer
from betty.project.data import ProjectConfiguration
from betty.render import RenderDispatcher, RendererDefinition
from betty.serde import SerializerDefinition, serializer_for
from betty.service.factory import DataManufacturable
from betty.service.level import UNIVERSE, ServiceLevel
from betty.service.provider import service

if TYPE_CHECKING:
    from collections.abc import (
        AsyncIterator,
        Collection,
        Iterator,
        MutableSequence,
        Sequence,
    )

    from babel import Locale

    from betty.app import App
    from betty.copyright_notice import CopyrightNotice
    from betty.jinja import Environment
    from betty.license import License
    from betty.plugin.manager import PluginManager
    from betty.url import UrlGenerator


@final
class Project(DataManufacturable[ProjectConfiguration], ServiceLevel, ManagedLifeCycle):
    """
    Define a Betty project.

    A project combines project configuration and the resulting ancestry.

    .. list-table::
       :widths: 10 20
       :header-rows: 0

       * - Configuration
         - :py:class:`betty.project.data.ProjectConfiguration`
    """

    def __init__(
        self,
        app: App,
        configuration_file: Path,
        /,
        *,
        configuration: ProjectConfiguration,
        ancestry: Ancestry | None = None,
        plugins: PluginManager | None = None,
    ):
        super().__init__(plugins=plugins)
        self.life_cycle.on_bootstrap(self._ensure_locale)
        self.life_cycle.on_bootstrap(self._validate)
        self._app = app
        self._configuration = configuration
        self._configuration_file = configuration_file
        self._ancestry = Ancestry() if ancestry is None else ancestry

    def _ensure_locale(self) -> None:
        if not self._configuration.locales:
            self._configuration.locales.add(DEFAULT_LOCALE)

    async def _validate(self) -> None:
        for entity_type in self._configuration.entity_types:
            await entity_type.validate(self)

    @override
    @classmethod
    def new_data_cls(cls) -> type[ProjectConfiguration]:
        return ProjectConfiguration

    @override
    @classmethod
    async def new(cls, services: ServiceLevel, data: ProjectConfiguration, /) -> Self:
        raise NotImplementedError(
            f"Creating a new {fully_qualified_name(cls)} from its configuration is not yet supported."
        )

    @property
    def configuration(self) -> ProjectConfiguration:
        """
        The project configuration.
        """
        return self._configuration

    @classmethod
    @asynccontextmanager
    async def new_isolated(
        cls,
        app: App,
        *,
        ancestry: Ancestry | None = None,
        configuration: ProjectConfiguration | None = None,
        configuration_file: Path | None = None,
        plugins: PluginManager | None = None,
    ) -> AsyncIterator[Self]:
        """
        Creat a new, isolated, temporary project.

        The project will not leave any traces on the system, except when it uses
        global Betty functionality such as caches.
        """
        async with AsyncExitStack() as stack:
            if configuration_file is None:
                configuration_file = (
                    Path(await stack.enter_async_context(TemporaryDirectory()))
                    / "betty.json"
                )
            yield cls(
                app,
                configuration_file,
                configuration=ProjectConfiguration(
                    title="Betty", url="https://example.com"
                )
                if configuration is None
                else configuration,
                ancestry=ancestry,
                plugins=plugins,
            )

    @property
    def configuration_file(self) -> Path:
        """
        The path to the configuration's file.
        """
        return self._configuration_file

    async def set_configuration_file(self, configuration_file: Path, /) -> None:
        """
        Set the path to the configuration's file.
        """
        if configuration_file == self._configuration_file:
            return
        serializer_for(
            list(await UNIVERSE.plugins.plugins(SerializerDefinition)),
            configuration_file.suffix,
        )
        self._configuration_file = configuration_file

    @property
    def directory(self) -> Path:
        """
        The project directory path.

        Betty will look for resources in this directory, and place generated artifacts there. It is expected
        that no other applications or projects share this same directory.
        """
        return self.configuration_file.parent

    @property
    def output_directory(self) -> Path:
        """
        The output directory path.
        """
        return self.directory / "output"

    @property
    def assets_directory(self) -> Path:
        """
        The :doc:`assets directory path </usage/assets>`.
        """
        return self.directory / "assets"

    @property
    def www_directory(self) -> Path:
        """
        The WWW directory path.
        """
        return self.output_directory / "www"

    def localize_www_directory(self, locale: Locale) -> Path:
        """
        Get the WWW directory path for a locale.
        """
        if self.configuration.multilingual:
            return self.www_directory / self.configuration.locales[locale].slug
        return self.www_directory

    @property
    def app(self) -> App:
        """
        The application this project is run within.
        """
        return self._app

    @property
    def name(self) -> MachineName:
        """
        The project name.

        If no project name was configured, this defaults to the hash of the configuration file path.
        """
        if self._configuration.name is None:
            return MachineName(hashid(str(self.configuration_file)))
        return self._configuration.name

    @property
    def ancestry(self) -> Ancestry:
        """
        The project's ancestry.
        """
        return self._ancestry

    @service
    async def _project_assets(self) -> AssetRepository:
        asset_paths = [self.assets_directory]
        extensions = await self.extensions
        for project_extension in extensions.flatten():
            extension_assets_directory_path = (
                project_extension.plugin().assets_directory
            )
            if extension_assets_directory_path is not None:
                asset_paths.append(extension_assets_directory_path)
        return StaticAssetRepository(*asset_paths)

    @service
    async def assets(self) -> AssetRepository:
        """
        The assets file system.
        """
        return ProxyAssetRepository(await self._project_assets, self.app.assets)

    @service
    async def translations(self) -> TranslationRepository:
        """
        The available translations.
        """
        return ProxyTranslationRepository(
            AssetTranslationRepository(
                await self._project_assets, self.app.binary_file_cache
            ),
            await self.app.translations,
        )

    @service
    async def localizers(self) -> LocalizerRepository:
        """
        The available localizers.
        """
        return LocalizerRepository(await self.translations)

    @service
    async def public_localizers(self) -> Collection[Localizer]:
        """
        The public localizers.
        """
        localizers = await self.localizers
        return [localizers.get(locale) for locale in self.configuration.locales.keys()]  # noqa: SIM118

    @service
    async def url_generator(self) -> UrlGenerator:
        """
        The URL generator.
        """
        from betty.project.url import new_project_url_generator

        return await new_project_url_generator(self)

    @service
    async def jinja(self) -> Environment:
        """
        The Jinja2 environment.
        """
        from betty.jinja import Environment

        return await Environment.new(self)

    @service
    async def renderer(self) -> RenderDispatcher:
        """
        The  content renderer.
        """
        return RenderDispatcher(
            *[
                await self.factory.new(plugin.cls)
                for plugin in await self.plugins.plugins(RendererDefinition)
            ]
        )

    @service
    async def extensions(self) -> ProjectExtensions:
        """
        The enabled extensions.
        """
        extensions = await self.plugins.plugins(ExtensionDefinition)
        configured_extension_definitions = []
        configured_extension_manufacturers = {}
        for extension_manufacturer in self.configuration.extensions:
            configured_extension_definitions.append(
                extensions[extension_manufacturer.plugin_id]
            )
            configured_extension_manufacturers[extension_manufacturer.plugin_id] = (
                extension_manufacturer
            )

        extensions_sorter = await sort_dependent_plugin_graph(
            extensions, configured_extension_definitions
        )
        extensions_sorter.prepare()

        theme_count = 0
        enabled_extensions = []
        while extensions_sorter.is_active():
            enabled_extension_ids_batch = extensions_sorter.get_ready()
            enabled_extension_batch: MutableSequence[Extension] = []
            for enabled_extension_id in enabled_extension_ids_batch:
                enabled_extension_definition = extensions[enabled_extension_id]
                if enabled_extension_definition.theme:
                    theme_count += 1
                if enabled_extension_id in configured_extension_manufacturers:
                    extension = await configured_extension_manufacturers[
                        enabled_extension_id
                    ](self)
                else:
                    extension = await self.factory.new(enabled_extension_definition.cls)
                await extension.bootstrap()
                enabled_extension_batch.append(extension)
                extensions_sorter.done(enabled_extension_id)
            self.life_cycle.attach(*enabled_extension_batch)
            enabled_extensions.append(
                sorted(
                    enabled_extension_batch,
                    key=lambda extension: extension.plugin().id,
                )
            )
        initialized_extensions = ProjectExtensions(enabled_extensions)

        # Users may not realize no theme is enabled, and be confused by their site looking bare.
        # Warn them out of courtesy.
        if theme_count == 0:
            await self.app.user.message_warning(
                _(
                    'Your project has no theme enabled. This means your site\'s pages may look bare. Try the "raspberry-mint" extension.'
                )
            )

        return initialized_extensions

    @property
    def logo(self) -> Path:
        """
        The path to the logo file.
        """
        return (
            self._configuration.logo
            or betty.dirs.ASSETS_DIRECTORY_PATH
            / "public"
            / "static"
            / "betty-512x512.png"
        )

    @service
    async def copyright_notice(self) -> CopyrightNotice:
        """
        The overall project copyright.
        """
        return await self.factory.new(self.configuration.copyright_notice)

    @service
    async def license(self) -> License:
        """
        The overall project license.
        """
        return await self.factory.new(self.configuration.license)

    @service
    def privatizer(self) -> Privatizer:
        """
        The privatizer.
        """
        return Privatizer(self.configuration.lifetime_threshold, user=self.app.user)

    async def new_document(
        self,
        resource: object = None,
        resource_url: object = None,
        **kwargs: object,
    ) -> Document:
        """
        Create a new document.
        """
        extensions = await self.extensions
        return Document(
            resource,
            resource_url,
            **{
                key: value
                for extension in extensions.flatten()
                if isinstance(extension, DocumentProvider)
                for (key, value) in extension.new_document_vars().items()
            },
            **kwargs,
        )


@final
class ProjectExtensions:
    """
    Manage the extensions running within the :py:class:`betty.project.Project`.
    """

    def __init__(self, project_extensions: Sequence[Sequence[Extension]]):
        super().__init__()
        self._project_extensions = project_extensions

    @overload
    def __getitem__[ExensionT: Extension](
        self, extension: type[ExensionT]
    ) -> ExensionT:
        pass

    @overload
    def __getitem__(self, extension: ResolvableId[ExtensionDefinition]) -> Extension:
        pass

    def __getitem__(self, extension: ResolvableId[ExtensionDefinition]) -> Extension:
        extension_id = resolve_id(extension)
        for project_extension in self.flatten():
            if project_extension.plugin().id == extension_id:
                return project_extension
        raise KeyError(f'Unknown extension of type "{extension_id}"')

    def __iter__(self) -> Iterator[Iterator[Extension]]:
        """
        Iterate over all extensions, in topologically sorted batches.

        Each item is a batch of extensions. Items are ordered because later items depend
        on earlier items. The extensions in each item do not depend on each other and their
        order has no meaning. However, implementations SHOULD sort the extensions in each
        item in a stable fashion for reproducability.
        """
        # Use a generator so we discourage calling code from storing the result.
        for batch in self._project_extensions:
            yield (project_extension for project_extension in batch)

    def flatten(self) -> Iterator[Extension]:
        """
        Get a sequence of topologically sorted extensions.
        """
        for batch in self:
            yield from batch

    def __contains__(self, extension: ResolvableId[ExtensionDefinition]) -> bool:
        if isinstance(extension, type) and issubclass(extension, Extension):
            extension = extension.plugin()
        try:
            self[resolve_id(extension)]
        except KeyError:
            return False
        else:
            return True



"""
Multiprocessing functionality.
"""

from concurrent import futures
from multiprocessing import get_context
from signal import SIG_IGN, SIGINT, signal

CONTEXT = get_context("spawn")


class ProcessPoolExecutor(futures.ProcessPoolExecutor):
    """
    Like :py:class:`concurrent.futures.ProcessPoolExecutor`, but with error handling and low memory consumption.

    This uses the ``spawn`` method to create new processes, which is the Python 3.14 default.

    This ignores ``SIGINT``/:py:class:`KeyboardInterrupt` delivered to it by the parent process to prevent unhelpful
    additional :py:class:`KeyboardInterrupt` being raised from within the process pool.
    """

    def __init__(
        self, max_workers: int | None = None, *, max_tasks_per_child: int | None = None
    ):
        """
        Initialize a new instance.
        """
        super().__init__(
            initializer=signal,
            initargs=(SIGINT, SIG_IGN),
            max_workers=max_workers,
            max_tasks_per_child=max_tasks_per_child,
            mp_context=CONTEXT,
        )



"""
Task progress management.
"""

from abc import ABC, abstractmethod


class Progress(ABC):
    """
    Track the progress of a number of tasks.
    """

    @abstractmethod
    async def add(self, add: int = 1, /) -> None:
        """
        Add a number of tasks to the total.
        """

    @abstractmethod
    async def done(self, done: int = 1, /) -> None:
        """
        Mark a number of tasks done.
        """



"""
Aggregate data types.
"""

from __future__ import annotations

from typing import Any

from betty.data import DataDefinition
from betty.data.indicator.selector import Element


class AggregateDefinition[DataClsT, ElementT: Element[Any] = Element[Any]](
    DataDefinition[DataClsT]
):
    """
    Define an aggregate data type, e.g. data that consists of other parts.
    """



"""
Defaults for the Raspberry Mint extension.
"""

from __future__ import annotations

from typing import TYPE_CHECKING

from betty.content import ContentManufacturer
from betty.content.contents import Box, BoxConfiguration, Notes
from betty.extension.maps.content import Attribution, Map
from betty.extension.raspberry_mint import SINGLE_COLUMN_TEXT_WIDTH
from betty.extension.raspberry_mint import ColorStyle as ColorStyleOption
from betty.extension.raspberry_mint.content import (
    Citations,
    ColorStyle,
    ColorStyleConfiguration,
    Columns,
    ColumnsConfiguration,
    Enclosees,
    ExternalLinks,
    Facts,
    Families,
    FileReferees,
    Media,
    MediaGallery,
    Presences,
    PresencesConfiguration,
    Section,
    SectionConfiguration,
    Timeline,
)
from betty.extension.trees.content import Tree
from betty.extension.wiki.content import WikipediaSummary
from betty.locale.localizable.gettext import _
from betty.locale.localizable.static import StaticTranslations
from betty.role.roles import Subject, Witness

if TYPE_CHECKING:
    from collections.abc import Collection

    from betty.extension.raspberry_mint.data import ResolvableRegionalContent
    from betty.locale.localizable import Localizable
    from betty.locale.localize import Localizer


def regional_content(*, localizers: Collection[Localizer]) -> ResolvableRegionalContent:
    """
    The default regional content configuration.
    """

    def _make_dumpable(localizable: Localizable) -> StaticTranslations:
        return StaticTranslations.resolve(localizable, localizers)

    return {
        "entity-page-content": [
            Media,
            ContentManufacturer(
                Section,
                SectionConfiguration(
                    ContentManufacturer(
                        Columns,
                        ColumnsConfiguration(
                            [[WikipediaSummary]], width=SINGLE_COLUMN_TEXT_WIDTH
                        ),
                    ),
                    heading=_make_dumpable(_("Wikipedia says...")),
                    name="wikipedia",
                ),
            ),
            ContentManufacturer(
                Box,
                BoxConfiguration(
                    Map, min_height="500px", height="75vh", max_height="1000px"
                ),
            ),
            ContentManufacturer(
                ColorStyle,
                ColorStyleConfiguration(
                    ContentManufacturer(Columns, ColumnsConfiguration([[Attribution]])),
                    style=ColorStyleOption.LIGHT_CONTRAST,
                ),
            ),
            ContentManufacturer(
                Columns,
                ColumnsConfiguration([[Enclosees]], width=SINGLE_COLUMN_TEXT_WIDTH),
            ),
            ContentManufacturer(
                Section,
                SectionConfiguration(
                    ContentManufacturer(
                        Columns,
                        ColumnsConfiguration([[Notes]], width=SINGLE_COLUMN_TEXT_WIDTH),
                    ),
                    heading=_make_dumpable(_("Notes")),
                    name="notes",
                ),
            ),
            ContentManufacturer(
                Section,
                SectionConfiguration(
                    ContentManufacturer(
                        Presences, PresencesConfiguration(include=[Subject])
                    ),
                    heading=_make_dumpable(_("Subjects")),
                    name="attendees-subject",
                ),
            ),
            ContentManufacturer(
                Section,
                SectionConfiguration(
                    ContentManufacturer(
                        Presences, PresencesConfiguration(include=[Witness])
                    ),
                    heading=_make_dumpable(_("Witnesses")),
                    name="attendees-witness",
                ),
            ),
            ContentManufacturer(
                Section,
                SectionConfiguration(
                    ContentManufacturer(
                        Presences,
                        PresencesConfiguration(exclude=[Subject, Witness]),
                    ),
                    heading=_make_dumpable(_("Other attendees")),
                    name="attendees-other",
                ),
            ),
            ContentManufacturer(
                Section,
                SectionConfiguration(
                    Families, heading=_make_dumpable(_("Family")), name="family"
                ),
            ),
            ContentManufacturer(
                Box,
                BoxConfiguration(
                    Tree, min_height="500px", height="75vh", max_height="1000px"
                ),
            ),
            ContentManufacturer(
                Section,
                SectionConfiguration(
                    ContentManufacturer(Columns, ColumnsConfiguration([[Timeline]])),
                    heading=_make_dumpable(_("Timeline")),
                    name="timeline",
                ),
            ),
            ContentManufacturer(
                Section,
                SectionConfiguration(
                    ContentManufacturer(
                        Columns,
                        ColumnsConfiguration([[Facts]], width=SINGLE_COLUMN_TEXT_WIDTH),
                    ),
                    heading=_make_dumpable(_("Facts")),
                    name="facts",
                ),
            ),
            ContentManufacturer(
                ColorStyle,
                ColorStyleConfiguration(
                    ContentManufacturer(
                        Section,
                        SectionConfiguration(
                            MediaGallery,
                            heading=_make_dumpable(_("Media")),
                            name="media",
                        ),
                    ),
                    style=ColorStyleOption.DARK,
                ),
            ),
            ContentManufacturer(
                Section,
                SectionConfiguration(
                    ContentManufacturer(
                        Columns,
                        ColumnsConfiguration(
                            [[FileReferees]], width=SINGLE_COLUMN_TEXT_WIDTH
                        ),
                    ),
                    heading=_make_dumpable(_("Appearances")),
                    name="appearances",
                ),
            ),
            ContentManufacturer(
                Section,
                SectionConfiguration(
                    ContentManufacturer(
                        Columns,
                        ColumnsConfiguration(
                            [[Citations]], width=SINGLE_COLUMN_TEXT_WIDTH
                        ),
                    ),
                    heading=_make_dumpable(_("Citations")),
                    name="citations",
                ),
            ),
            ContentManufacturer(
                Section,
                SectionConfiguration(
                    ContentManufacturer(
                        Columns,
                        ColumnsConfiguration(
                            [[ExternalLinks]], width=SINGLE_COLUMN_TEXT_WIDTH
                        ),
                    ),
                    heading=_make_dumpable(_("External links")),
                    name="external-links",
                ),
            ),
        ]
    }



"""
Project data.
"""

from __future__ import annotations

from typing import TYPE_CHECKING, final
from urllib.parse import urlparse

from babel import Locale

from betty.ancestry.person import Person
from betty.assertion import assert_number
from betty.collection.keyed.adapter import MutableKeyedCollectionAdapter
from betty.copyright_notice import (
    CopyrightNotice,
    CopyrightNoticeDefinition,
    CopyrightNoticeManufacturer,
)
from betty.copyright_notice.data import CopyrightNoticeDefinitionConfiguration
from betty.data import Data, Sample
from betty.data.aggregate.collection.keyed import KeyedCollectionDefinition
from betty.data.aggregate.record.object import AttrDefinition, ObjectDefinition
from betty.data.bool import BoolDefinition
from betty.data.indicator.selector import Attr
from betty.data.int import IntDefinition
from betty.data.str import StrDefinition
from betty.dirs import ASSETS_DIRECTORY_PATH
from betty.event_type import EventTypeDefinition
from betty.event_type.data import EventTypeDefinitionConfiguration
from betty.exception import HumanFacingException
from betty.extension import ExtensionManufacturer
from betty.gender import GenderDefinition
from betty.gender.data import GenderDefinitionConfiguration
from betty.license import License, LicenseDefinition, LicenseManufacturer
from betty.license.data import LicenseDefinitionConfiguration
from betty.locale import (
    DEFAULT_LOCALE,
    ResolvableLocale,
    resolve_locale,
    to_language_tag,
)
from betty.locale.data import LocaleProperty
from betty.locale.localizable.gettext import _
from betty.locale.localizable.property import LocalizableProperty
from betty.machine_name import MachineName, MachineNameProperty, ResolvableMachineName
from betty.model import EntityDefinition
from betty.pathlib import FilePathDefinition
from betty.place_type import PlaceTypeDefinition
from betty.place_type.data import PlaceTypeDefinitionConfiguration
from betty.plugin import ResolvableId, resolve_id
from betty.plugin.data.property import PluginDefinitionConfigurationsProperty
from betty.project import Extension, ExtensionDefinition
from betty.property import (
    Optional,
    Property,
)
from betty.property.collection.keyed import KeyedCollectionProperty
from betty.role import RoleDefinition
from betty.role.data import RoleDefinitionConfiguration
from betty.sample import Size

if TYPE_CHECKING:
    from collections.abc import Iterable
    from pathlib import Path

    from betty.locale.localizable import ResolvableLocalizable
    from betty.plugin.factory import (
        ResolvablePluginManufacturer,
        ResolvablePluginManufacturerSequence,
    )
    from betty.service.level import ServiceLevel

DEFAULT_LIFETIME_THRESHOLD = 123
"""
The default age by which people are presumed dead.

This is based on `Jeanne Louise Calment <https://www.guinnessworldrecords.com/world-records/oldest-person/>`_ who is
the oldest verified person to ever have lived.
"""


@final
@ObjectDefinition(
    label=_("Entity type configuration"),
    samples=[
        lambda: Sample(
            EntityTypeConfiguration(entity_type=Person),
            label="Minimal",
            size=Size.MINIMAL,
        ),
        lambda: Sample(
            EntityTypeConfiguration(entity_type=Person, generate_html_list=False),
            label="Full",
            size=Size.FULL,
        ),
    ],
)
class EntityTypeConfiguration(Data[ObjectDefinition["EntityTypeConfiguration"]]):
    """
    Configure a single entity type for a project.

    .. data:: betty.project.data:EntityTypeConfiguration
    """

    def __init__(
        self,
        *,
        entity_type: ResolvableId[EntityDefinition],
        generate_html_list: bool = True,
    ):
        self._entity_type = resolve_id(entity_type)
        self.generate_html_list = generate_html_list

    @property
    @AttrDefinition(MachineName)
    def entity_type(self) -> MachineName:
        """
        The ID of the configured entity type.
        """
        return self._entity_type

    @property
    @AttrDefinition(
        BoolDefinition(label=_("Generate list HTML page")),
        omit_load=True,
        omit_dump=lambda data: data is True,
    )
    def generate_html_list(self) -> bool:
        """
        Whether to generate listing web pages for entities of this type.
        """
        return self._generate_html_list

    @generate_html_list.setter
    def generate_html_list(self, generate_html_list: bool) -> None:
        self._generate_html_list = generate_html_list

    async def validate(self, services: ServiceLevel, /) -> None:
        """
        Validate the configuration.
        """
        entity_type = (await services.plugins.plugins(EntityDefinition)).get(
            self._entity_type
        )
        if self.generate_html_list and not entity_type.public_facing:
            raise HumanFacingException(
                _(
                    "Cannot generate pages for {entity_type}, because it is not a public-facing entity type."
                ).format(entity_type=entity_type.label)
            )


@final
@ObjectDefinition(
    label=_("Project locale"),
    samples=[
        lambda: Sample(
            ProjectLocale(Locale("nl", "NL")), label="Minimal", size=Size.MINIMAL
        ),
        lambda: Sample(
            ProjectLocale(Locale("nl", "NL"), alias="nl"),
            label="Full",
            size=Size.FULL,
        ),
    ],
)
class ProjectLocale(Data["ObjectDefinition"]):
    """
    A locale to use for a project.

    .. data:: betty.project.data:ProjectLocale
    """

    locale = LocaleProperty()
    """
    The locale.
    """

    @staticmethod
    def _resolve_alias(alias: str) -> str:
        if "/" in alias:
            raise HumanFacingException(_("Locale aliases must not contain slashes."))
        return alias

    alias = Optional(Property(StrDefinition(label=_("Alias")), resolver=_resolve_alias))
    """
    A shorthand alias to use instead of the full language tag, such as when rendering URLs.
    """

    def __init__(self, /, locale: ResolvableLocale, *, alias: str | None = None):
        super().__init__()
        self.locale = locale
        self.alias = alias

    @property
    def slug(self) -> str:
        """
        The URL slug.
        """
        if self.alias is None:
            return to_language_tag(self.locale)
        return self.alias


@final
@ObjectDefinition(
    label=_("Project configuration"),
    samples=[
        lambda: Sample(
            ProjectConfiguration(title="Betty", url="https://example.com"),
            label="Minimal",
            size=Size.MINIMAL,
        ),
        lambda: Sample(
            ProjectConfiguration(
                author="Bart Feenstra",
                clean_urls=True,
                copyright_notice=CopyrightNoticeManufacturer.data()
                .samples.get(Size.FULL)
                .subject,
                copyright_notices=[
                    CopyrightNoticeDefinitionConfiguration.data()
                    .samples.get(Size.FULL)
                    .subject
                ],
                debug=True,
                entity_types=[
                    EntityTypeConfiguration.data().samples.get(Size.FULL).subject
                ],
                event_types=[
                    EventTypeDefinitionConfiguration.data()
                    .samples.get(Size.FULL)
                    .subject
                ],
                genders=[
                    GenderDefinitionConfiguration.data().samples.get(Size.FULL).subject
                ],
                logo=ASSETS_DIRECTORY_PATH / "public" / "static" / "betty-512x512.png",
                license=LicenseManufacturer.data().samples.get(Size.FULL).subject,
                licenses=[
                    LicenseDefinitionConfiguration.data().samples.get(Size.FULL).subject
                ],
                lifetime_threshold=123,
                locales=[ProjectLocale.data().samples.get(Size.FULL).subject],
                name="betty-ancestry",
                place_types=[
                    PlaceTypeDefinitionConfiguration.data()
                    .samples.get(Size.FULL)
                    .subject
                ],
                roles=[
                    RoleDefinitionConfiguration.data().samples.get(Size.FULL).subject
                ],
                title="Betty's ancestry",
                url="https://ancestry.example.com/betty",
            ),
            label="Full",
            size=Size.FULL,
        ),
    ],
)
class ProjectConfiguration(Data):
    """
    Configuration for a :py:class:`betty.project.Project`.

    .. data:: betty.project.data:ProjectConfiguration
    """

    author = Optional(LocalizableProperty(label=_("Author")))
    """
    The project's author.
    """

    clean_urls = Property(
        BoolDefinition(
            label=_("Clean URLs"),
            description=_(
                'Whether to use clean URLs: "/path" instead of "/path/index.html".'
            ),
        ),
        omit_load=True,
        omit_dump=lambda data: data is False,
    )
    """
    Whether to generate clean URLs such as ``/person/first-person`` instead of ``/person/first-person/index.html``.

    Generated artifacts will require web server that supports this.
    """

    copyright_notice = Property(
        CopyrightNoticeManufacturer,
        omit_load=True,
        omit_dump=lambda data: data == ProjectConfiguration._default_copyright_notice(),
        default=lambda: ProjectConfiguration._default_copyright_notice(),
        resolver=CopyrightNoticeManufacturer.resolve,
    )
    """
    The project-wide copyright notice.
    """

    copyright_notices = PluginDefinitionConfigurationsProperty(
        CopyrightNoticeDefinition, CopyrightNoticeDefinitionConfiguration
    )
    """
    The :py:class:`betty.copyright_notice.CopyrightNotice` plugins created by this project.
    """

    debug = Property(
        BoolDefinition(
            label=_("Debugging mode"),
            description=_(
                "Whether to output more detailed logs and disable optimizations that make debugging harder."
            ),
        ),
        omit_load=True,
        omit_dump=lambda data: data is False,
    )
    """
    Whether to enable debugging for project jobs.

    This setting is disabled by default.

    Enabling this generally results in:

    - More verbose logging output
    - job artifacts (e.g. generated sites)
    """

    entity_types = KeyedCollectionProperty(
        KeyedCollectionDefinition(
            value=EntityTypeConfiguration,
            label=_("Entity types"),
            key=Attr("entity_type"),
            factory=lambda: MutableKeyedCollectionAdapter(
                key=lambda item: item.entity_type,
                value_resolver=lambda data: (
                    data
                    if isinstance(data, EntityTypeConfiguration)
                    else EntityTypeConfiguration(entity_type=data)
                ),
            ),
        ),
        omit_load=True,
        omit_dump=lambda data: not data,
    )
    """
    The available entity types.
    """

    event_types = PluginDefinitionConfigurationsProperty(
        EventTypeDefinition, EventTypeDefinitionConfiguration
    )
    """
    The :py:class:`betty.event_type.EventType` plugins created by this project.
    """

    extensions = KeyedCollectionProperty(
        KeyedCollectionDefinition(
            value=ExtensionManufacturer,
            label=ExtensionDefinition.type().label_plural,
            key=Attr("plugin_id"),
            factory=lambda: MutableKeyedCollectionAdapter(
                key=lambda data: data.plugin_id,
                key_resolver=resolve_id,
                value_resolver=ExtensionManufacturer.resolve,
            ),
        ),
        omit_load=True,
        omit_dump=lambda data: not data,
    )
    """
    The extensions to enable for the project.
    """

    genders = PluginDefinitionConfigurationsProperty(
        GenderDefinition, GenderDefinitionConfiguration
    )
    """
    The :py:class:`betty.gender.Gender` plugins created by this project.
    """

    license = Property(
        LicenseManufacturer,
        omit_load=True,
        omit_dump=lambda data: data == ProjectConfiguration._default_license(),
        default=lambda: ProjectConfiguration._default_license(),
        resolver=LicenseManufacturer.resolve,
    )
    """
    The project-wide license.
    """

    licenses = PluginDefinitionConfigurationsProperty(
        LicenseDefinition, LicenseDefinitionConfiguration
    )
    """
    The :py:class:`betty.license.License` plugins created by this project.
    """

    lifetime_threshold = Property(
        IntDefinition(
            label=_("Lifetime threshold"),
            description=_(
                "The number of years people are expected to live at most, e.g. after which they are presumed to have died."
            ),
        ),
        omit_load=True,
        omit_dump=lambda data: data == DEFAULT_LIFETIME_THRESHOLD,
        resolver=assert_number(minimum=1),
    )
    """
    The lifetime threshold indicates when people are considered dead.

    This setting defaults to :py:const:`betty.project.data.DEFAULT_LIFETIME_THRESHOLD`.

    The value is an integer expressing the age in years over which people are
    presumed to have died.
    """

    locales = KeyedCollectionProperty(
        KeyedCollectionDefinition(
            value=ProjectLocale,
            label=_("Locales"),
            key=Attr("locale"),
            order_dump=True,
            factory=lambda: MutableKeyedCollectionAdapter(
                key=lambda item: item.locale,
                key_resolver=resolve_locale,
                value_resolver=lambda value: (
                    value
                    if isinstance(value, ProjectLocale)
                    else ProjectLocale(resolve_locale(value))
                ),
            ),
        ),
        omit_load=True,
        omit_dump=lambda data: not data,
        default=lambda: [DEFAULT_LOCALE],
    )
    """
    The configured locales.
    """

    logo = Optional(Property(FilePathDefinition(), label=_("Logo")))
    """
    The project logo.
    """

    name = Optional(MachineNameProperty())
    """
    The project's machine name.
    """

    place_types = PluginDefinitionConfigurationsProperty(
        PlaceTypeDefinition, PlaceTypeDefinitionConfiguration
    )
    """
    The :py:class:`betty.place_type.PlaceType` plugins created by this project.
    """

    roles = PluginDefinitionConfigurationsProperty(
        RoleDefinition, RoleDefinitionConfiguration
    )
    """
    The :py:class:`betty.role.Role` plugins created by this project.
    """

    title = LocalizableProperty(label=_("Title"))
    """
    The human-readable project title.
    """

    def __init__(
        self,
        *,
        title: ResolvableLocalizable,
        url: str,
        author: ResolvableLocalizable | None = None,
        clean_urls: bool = False,
        copyright_notice: ResolvablePluginManufacturer[
            CopyrightNoticeDefinition, CopyrightNotice
        ]
        | None = None,
        copyright_notices: Iterable[CopyrightNoticeDefinitionConfiguration]
        | None = None,
        debug: bool = False,
        entity_types: Iterable[EntityTypeConfiguration | ResolvableId[EntityDefinition]]
        | None = None,
        event_types: Iterable[EventTypeDefinitionConfiguration] | None = None,
        extensions: ResolvablePluginManufacturerSequence[ExtensionDefinition, Extension]
        | None = None,
        genders: Iterable[GenderDefinitionConfiguration] | None = None,
        license: ResolvablePluginManufacturer[LicenseDefinition, License] | None = None,  # noqa: A002
        licenses: Iterable[LicenseDefinitionConfiguration] | None = None,
        lifetime_threshold: int = DEFAULT_LIFETIME_THRESHOLD,
        locales: Iterable[ResolvableLocale | ProjectLocale] | None = None,
        logo: Path | None = None,
        name: ResolvableMachineName | None = None,
        place_types: Iterable[PlaceTypeDefinitionConfiguration] | None = None,
        roles: Iterable[RoleDefinitionConfiguration] | None = None,
    ):
        super().__init__()
        self.author = author
        self.clean_urls = clean_urls
        if copyright_notice is not None:
            self.copyright_notice = CopyrightNoticeManufacturer.resolve(
                copyright_notice
            )
        if copyright_notices is not None:
            self.copyright_notices = copyright_notices
        self.debug = debug
        if entity_types is not None:
            self.entity_types = entity_types
        if event_types is not None:
            self.event_types = event_types
        if extensions is not None:
            self.extensions = extensions  # ty:ignore[invalid-assignment]
        if genders is not None:
            self.genders = genders
        if license is not None:
            self.license = LicenseManufacturer.resolve(license)
        if licenses is not None:
            self.licenses = licenses
        self.lifetime_threshold = lifetime_threshold
        self.logo = logo
        if locales is not None:
            self.locales = locales
        self.name = name
        if place_types is not None:
            self.place_types = place_types
        if roles is not None:
            self.roles = roles
        self.title = title
        self.url = url

    @classmethod
    def _default_copyright_notice(cls) -> CopyrightNoticeManufacturer:
        from betty.copyright_notice.copyright_notices import ProjectAuthor

        return CopyrightNoticeManufacturer(ProjectAuthor)

    @classmethod
    def _default_license(cls) -> LicenseManufacturer:
        from betty.license.licenses import AllRightsReserved

        return LicenseManufacturer(AllRightsReserved)

    @property
    @AttrDefinition(
        StrDefinition(
            label=_("URL"),
            description=_(
                "The absolute, public URL at which the site will be published."
            ),
        )
    )
    def url(self) -> str:
        """
        The project's public URL.
        """
        return self._url

    @url.setter
    def url(self, url: str) -> None:
        url_parts = urlparse(url)
        if not url_parts.scheme:
            raise HumanFacingException(
                _("The URL must start with a scheme such as https:// or http://.")
            )
        if not url_parts.netloc:
            raise HumanFacingException(_("The URL must include a host."))
        self._url = f"{url_parts.scheme}://{url_parts.netloc}{url_parts.path}"

    @property
    def base_url(self) -> str:
        """
        The project's public URL's base URL.

        If the public URL is ``https://example.com``, the base URL is ``https://example.com``.
        If the public URL is ``https://example.com/my-ancestry-site``, the base URL is ``https://example.com``.
        If the public URL is ``https://my-ancestry-site.example.com``, the base URL is ``https://my-ancestry-site.example.com``.
        """
        url_parts = urlparse(self.url)
        return f"{url_parts.scheme}://{url_parts.netloc}"

    @property
    def root_path(self) -> str:
        """
        The project's public URL's root path.

        If the public URL is ``https://example.com``, the root path is an empty string.
        If the public URL is ``https://example.com/my-ancestry-site``, the root path is ``/my-ancestry-site``.
        """
        return urlparse(self.url).path.rstrip("/")

    @property
    def default_locale(self) -> ProjectLocale:
        """
        The default locale.
        """
        return next(iter(self.locales))

    @property
    def multilingual(self) -> bool:
        """
        Whether the configuration is multilingual.
        """
        return len(self.locales) > 1
